import re
import logging

ROUTES = []
WORKERS = []

logger = logging.getLogger(__name__)


def make_route(function, method, path, check_session, check_key):
    splitpath = []
    elts = path.split(b'/')
    pos = 0
    root = None
    # Seeking for regexp in the path, because we need to compile it.
    for elt in elts:
        if len(elt) == 0:
            continue

        # python 3
        if type(elt) is bytes and type(elt) is not str and \
           chr(elt[0]) == '(' and chr(elt[-1]) == ')':
            elt = elt.decode('utf-8')

        # regexp have to start with a ( end is terminated by a ).
        if type(elt) is str and elt[0] == '(' and elt[-1] == ')':
            # Append it compiled.
            splitpath.append(re.compile(elt))
        # string case
        else:
            if pos == 0:
                root = elt
            splitpath.append(elt)
        pos += 1
    # A path can't start with a regexp.
    if root is None:
        raise Exception("Wrong route format.")
    return dict(
        http_method=method,
        root=root,
        path=path,
        splitpath=splitpath,
        module=function.__module__,
        function=function.__name__,
        check_session=check_session,
        check_key=check_key,
    )


def add_route(method, path, check_session=True, check_key=False):
    """
    Function decorator for HTTP method/path -> API function mapping.
    """
    def func_wrapper(function):
        global ROUTES
        ROUTES.append(make_route(function, method, path, check_session,
                                 check_key))
        return function
    return func_wrapper


def get_routes():
    """
    Returns the routes.
    """
    return ROUTES


class Router:
    # Adapter to manage routes with an instance rather than a global.

    def add(self, routes):
        global ROUTES
        for route in routes:
            if route not in ROUTES:
                ROUTES.append(route)

    def remove(self, routes):
        global ROUTES
        for route in routes:
            if route in ROUTES:
                ROUTES.remove(route)


class RouteSet(list):
    def __init__(self, prefix=b''):
        self.prefix = prefix

    def delete(self, path, check_session=True, check_key=False):
        def register_route(f):
            self.append(
                make_route(f, 'DELETE', self.prefix + path, check_session,
                           check_key))
            return f
        return register_route

    def get(self, path, check_session=True, check_key=False):
        def register_route(f):
            self.append(
                make_route(f, 'GET', self.prefix + path, check_session,
                           check_key))
            return f
        return register_route

    def post(self, path, check_session=True, check_key=False):
        def register_route(f):
            self.append(
                make_route(f, 'POST', self.prefix + path, check_session,
                           check_key))
            return f
        return register_route



# flake8: noqa
# pytype: disable=import-error

import sys


PY3 = sys.version_info > (3,)
PY2 = not PY3

if PY2:
    import ConfigParser as configparser
    from UserDict import IterableUserDict
    from Queue import Empty
else:
    import configparser
    from collections import UserDict as IterableUserDict
    from queue import Empty



from argparse import (
    ArgumentParser, SUPPRESS as UNDEFINED_ARGUMENT, _VersionAction,
)
from socket import getfqdn
import logging
import os
import datetime
import getpass
import sys
from platform import python_version
from textwrap import dedent

from ..toolkit import taskmanager

from ..cli import Application
from ..toolkit.configuration import OptionSpec
from ..toolkit.versions import (
    format_pq_version,
    read_distinfo,
    read_libpq_version,
)
from ..daemon import daemonize
from ..httpd import HTTPDService
from ..routing import Router
from ..toolkit import validators as v
from ..toolkit.app import define_core_arguments
from ..toolkit.proctitle import ProcTitleManager
from ..toolkit.services import ServicesManager
from ..toolkit.tasklist.sqlite3_engine import TaskListSQLite3Engine
from ..notification import NotificationMgmt

logger = logging.getLogger('temboardagent.scripts.agent')


class VersionAction(_VersionAction):
    fmt = dedent("""\
    temBoard agent %(temboard)s
    System %(distname)s %(distversion)s
    Python %(python)s (%(pythonbin)s)
    psycopg2 %(psycopg2)s libpq %(libpq)s
    """)

    def __call__(self, parser, *_):
        print((self.fmt % self.inspect_versions()).strip())
        parser.exit()

    @classmethod
    def inspect_versions(cls):
        from psycopg2 import __version__ as psycopg2_version

        distinfos = read_distinfo()

        return dict(
            temboard=Application.VERSION,
            psycopg2=psycopg2_version,
            python=python_version(),
            pythonbin=sys.executable,
            distname=distinfos['NAME'],
            distversion=distinfos['VERSION'],
            libpq=format_pq_version(read_libpq_version()),
        )


def define_arguments(parser):
    define_core_arguments(parser)
    parser.add_argument(
        '-V', '--version',
        action=VersionAction,
        help='show version and exit',
    )
    parser.add_argument(
        '-d', '--daemon',
        action='store_true', dest='temboard_daemonize',
        help="Run in background.",
    )
    parser.add_argument(
        '-p', '--pid-file',
        action='store', dest='temboard_pidfile',
        help="PID file.",
    )


def list_options_specs():
    # Generate each option specs.
    section = 'temboard'
    yield OptionSpec(section, 'daemonize', default=False)
    yield OptionSpec(section, 'pidfile', default='/run/temboard-agent.pid')
    yield OptionSpec(
        section, 'address', default='0.0.0.0', validator=v.address)
    yield OptionSpec(section, 'port', validator=v.port, default=2345)
    yield OptionSpec(
        section, 'ssl_cert_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(
        section, 'ssl_key_file',
        default=OptionSpec.REQUIRED, validator=v.file_)
    yield OptionSpec(section, 'ssl_ca_cert_file', validator=v.file_)
    yield OptionSpec(section, 'key')
    yield OptionSpec(
        section, 'users',
        default='/etc/temboard-agent/users', validator=v.file_,
    )
    yield OptionSpec(section, 'hostname', default=getfqdn())
    home = os.environ.get('HOME', '/var/lib/temboard-agent')
    yield OptionSpec(section, 'home', default=home, validator=v.writeabledir)


class AgentApplication(Application):
    PROGRAM = "temboard-agent"

    def bootstrap_plugins(self):
        for plugin_name, plugin in self.plugins.items():
            if hasattr(plugin, 'bootstrap'):
                logger.debug("Boostraping plugin %s", plugin_name)
                plugin.bootstrap()

    def main(self, argv, environ):
        parser = ArgumentParser(
            prog='temboard-agent',
            description="temBoard agent.",
            argument_default=UNDEFINED_ARGUMENT,
        )
        define_arguments(parser)
        args = parser.parse_args(argv)

        setproctitle = ProcTitleManager(prefix='temboard-agent: ')
        setproctitle.setup()

        self.router = Router()

        task_queue = taskmanager.Queue()
        event_queue = taskmanager.Queue()

        self.worker_pool = taskmanager.WorkerPoolService(
            app=self, setproctitle=setproctitle, name='worker pool',
            task_queue=task_queue, event_queue=event_queue)
        self.services.append(self.worker_pool)

        self.scheduler = taskmanager.SchedulerService(
            app=self, setproctitle=setproctitle, name='scheduler',
            task_queue=task_queue, event_queue=event_queue)
        self.services.append(self.scheduler)

        self.bootstrap(args=args, environ=environ)
        self.log_versions()
        config = self.config

        # TaskList engine setup must be done before we load the plugins
        self.scheduler.task_list_engine = TaskListSQLite3Engine(
            os.path.join(config.temboard['home'], 'agent_tasks.db')
        )

        self.apply_config()

        if config.postgresql.instance:
            setproctitle.prefix += config.postgresql.instance + ': '

        # Run temboard-agent as a background daemon.
        if (config.temboard.daemonize):
            daemonize(config.temboard.pidfile)

        logger.info("Starting main process.")

        self.start_datetime = datetime.datetime.now()
        self.reload_datetime = None
        self.pid = os.getpid()
        self.user = getpass.getuser()

        # Bootstraping plugins
        self.bootstrap_plugins()
        # Boostraping action logs table
        NotificationMgmt.bootstrap(config)

        # Purge all legacy data queues
        home = config.temboard['home']
        if os.path.exists(home):
            [os.remove(os.path.join(home, f))
             for f in os.listdir(home) if f.endswith('.q')]

        services = ServicesManager()
        services.add(self.worker_pool)
        services.add(self.scheduler)

        with services:
            httpd = HTTPDService(
                self, setproctitle=setproctitle, name='main process',
                services=services)
            httpd.run()

        return 0

    def reload(self):
        super().reload()
        self.reload_datetime = datetime.datetime.now()


main = AgentApplication(specs=list_options_specs())

if __name__ == '__main__':  # pragma: no cover
    main()



from argparse import ArgumentParser, SUPPRESS as UNDEFINED_ARGUMENT
from sys import stdout
from getpass import getpass

from ..cli import Application
from ..usermgmt import hash_password
from ..errors import ConfigurationError, HTTPError, UserError
from ..usermgmt import get_user
from ..types import T_PASSWORD, T_USERNAME
from ..toolkit.app import define_core_arguments
from ..tools import validate_parameters
from .agent import list_options_specs


def ask_password():
    raw_pass1 = getpass("Password: ")
    raw_pass2 = getpass("Retype password: ")
    if raw_pass1 != raw_pass2:
        stdout.write("Sorry, passwords do not match.\n")
        return ask_password()
    try:
        password = raw_pass1
        validate_parameters({'password': password},
                            [('password', T_PASSWORD, False)])
    except HTTPError:
        stdout.write("Invalid password.\n")
        return ask_password()
    return password


def ask_username(config):
    raw_username = input("Username: ")
    try:
        get_user(config.temboard['users'], raw_username)
    except HTTPError:
        pass
    except ConfigurationError:
        pass
    else:
        stdout.write("User already exists.\n")
        return ask_username(config)
    try:
        username = raw_username
        validate_parameters({'username': username},
                            [('username', T_USERNAME, False)])
    except HTTPError:
        stdout.write("Invalid username.\n")
        return ask_username(config)
    return username


class AddUserApplication(Application):
    PROGRAM = "temboard-agent-adduser"

    def main(self, argv, environ):
        parser = ArgumentParser(
            prog='temboard-agent-adduser',
            description="Add a new temboard-agent user.",
            argument_default=UNDEFINED_ARGUMENT,
        )
        define_core_arguments(parser, appversion=Application.VERSION)
        args = parser.parse_args(argv)
        self.bootstrap(args=args, environ=environ)

        # Load configuration from the configuration file.
        username = ask_username(self.config)
        password = ask_password()
        hash_ = hash_password(username, password).decode('utf-8')
        try:
            with open(self.config.temboard.users, 'a') as fd:
                fd.write("{}:{}\n".format(username, hash_))
        except OSError as e:
            raise UserError(str(e))
        else:
            stdout.write("Done.\n")


main = AddUserApplication(specs=list_options_specs(), with_plugins=False)

if __name__ == '__main__':
    main()



import os
from sys import stdout
from getpass import getpass
import re
import json
import logging

from ..cli import Application
from ..errors import (
    HTTPError,
    UserError,
)
from ..types import T_PASSWORD, T_USERNAME
from ..toolkit.app import define_core_arguments
from ..tools import validate_parameters
from ..httpsclient import https_request
from .agent import list_options_specs


try:
    input = raw_input
except NameError:
    pass

logger = logging.getLogger(__name__)


def ask_password():
    try:
        raw_pass = os.environ['TEMBOARD_UI_PASSWORD']
    except KeyError:
        raw_pass = getpass(" Password: ")

    try:
        password = raw_pass
        validate_parameters({'password': password},
                            [('password', T_PASSWORD, False)])
    except HTTPError:
        stdout.write("Invalid password.\n")
        return ask_password()
    return password


def ask_username():
    try:
        raw_username = os.environ['TEMBOARD_UI_USER']
    except KeyError:
        raw_username = input(" Username: ")

    try:
        username = raw_username
        validate_parameters({'username': username},
                            [('username', T_USERNAME, False)])
    except HTTPError:
        stdout.write("Invalid username.\n")
        return ask_username()
    return username


class RegisterApplication(Application):
    """Register a couple PostgreSQL instance/agent to a temBoard UI."""

    PROGRAM = "temboard-agent-register"

    def main(self, argv, environ):
        parser = self.create_parser(add_help=False)
        self.define_arguments(parser)
        args = parser.parse_args(argv)
        self.bootstrap(args=args, environ=environ)
        return wrapped_main(args, self)

    def define_arguments(self, parser):
        define_core_arguments(parser, appversion=Application.VERSION)

        parser.add_argument(
            '-?', '--help',
            action='help',
            help='show this help message and exit')

        parser.add_argument(
            '-h', '--host',
            dest='host',
            help="Agent address. Default: %(default)s",
            default='localhost'
        )
        self.config_specs['temboard_port'].add_argument(
            parser, '-p', '--port',
            help="Agent listening TCP port. Default: %(default)s",
        )

        parser.add_argument(
            '-g', '--groups',
            dest='groups',
            help="Instance groups list, comma separated. Default: %(default)s",
            default=None,
        )
        parser.add_argument(
            'ui_address',
            metavar='TEMBOARD-UI-ADDRESS',
            help="temBoard UI address to register to.",
        )


def wrapped_main(args, app):
    discover_url = "https://{}:{}/discover".format(
        args.host, app.config.temboard.port)
    try:
        # Getting system/instance informations using agent's discovering API
        logger.info(
            "Getting system & PostgreSQL informations from the agent (%s) ...",
            discover_url)
        (code, content, cookies) = https_request(
            None,
            'GET',
            discover_url,
            headers={
                "Content-type": "application/json"
            }
        )
        infos = json.loads(content.decode("utf-8"))

        print("Login at %s ..." % (args.ui_address))
        username = ask_username()
        password = ask_password()
        (code, content, cookies) = https_request(
            None,
            'POST',
            "%s/json/login" % (args.ui_address.rstrip('/')),
            headers={
                "Content-type": "application/json"
            },
            data={'username': username, 'password': password}
        )
        temboard_cookie = None
        for cookie in cookies.split("\n"):
            cookie_content = cookie.split(";")[0]
            if re.match(r'^temboard=.*$', cookie_content):
                temboard_cookie = cookie_content
                continue

        if args.groups:
            groups = [g for g in args.groups.split(',')]
        else:
            groups = None

        # POSTing new instance
        print("Registering instance/agent to %s ..." % (args.ui_address))
        (code, content, cookies) = https_request(
            None,
            'POST',
            "%s/json/register/instance" % (args.ui_address.rstrip('/')),
            headers={
                "Content-type": "application/json",
                "Cookie": temboard_cookie
            },
            data={
                'hostname': infos['hostname'],
                'agent_key': app.config.temboard['key'],
                'agent_address': args.host,
                'agent_port': str(app.config.temboard['port']),
                'cpu': infos['cpu'],
                'memory_size': infos['memory_size'],
                'pg_port': infos['pg_port'],
                'pg_data': infos['pg_data'],
                'pg_version': infos['pg_version'],
                'pg_version_summary': infos['pg_version_summary'],
                'plugins': infos['plugins'],
                'groups': groups
            }
        )
        if code != 200:
            raise HTTPError(code, content)
        print("Done.")
    except UserError:
        raise
    except HTTPError as e:
        err = json.loads(e.read())
        raise UserError(err['error'])
    except Exception as e:
        raise UserError(str(e) or repr(e))

    return 0


main = RegisterApplication(specs=list_options_specs(), with_plugins=False)

if __name__ == '__main__':  # pragma: no cover
    main()



# This module provide an undocumented feature mainly for development purpose:
# execute a task foreground.
#
# Usage: python -m temboardagent.scripts.runtask [options] worker_name [arg
# ...]
#
# Note that our home-made background task implementation, we use a different
# semantic that state of the art background task implemtation:
#
# a task function is called a worker
# a worker process is called workerpool
# a message is called a task
# the message broker is called taskmanager
#
import logging.config
import sys
from argparse import (
    ArgumentParser,
    SUPPRESS as UNDEFINED_ARGUMENT,
)
from ast import literal_eval

from .agent import list_options_specs
from ..cli import Application
from ..notification import NotificationMgmt
from ..routing import Router
from ..toolkit.app import define_core_arguments
from ..toolkit.errors import UserError
from ..toolkit import taskmanager
from ..toolkit.tasklist.sqlite3_engine import TaskListSQLite3Engine


# Avoid using `__name__` which could be `__main__`
module_name = 'temboardagent.scripts.runtask'

logger = logging.getLogger(module_name)


class TaskApplication(Application):
    PROGRAM = "temboardagent.scripts.runtask"

    def main(self, argv, environ):

        # Required to register plugins.
        self.router = Router()
        self.worker_pool = taskmanager.WorkerPoolService(
            app=self, name='worker pool',
            task_queue=None, event_queue=None)
        self.services.append(self.worker_pool)
        self.scheduler = taskmanager.SchedulerService(
            app=self, name='scheduler',
            task_queue=None, event_queue=None)
        self.scheduler.task_list_engine = TaskListSQLite3Engine(":memory:")
        self.services.append(self.scheduler)

        # C O N F I G U R A T I O N

        parser = ArgumentParser(
            prog=module_name,
            description="Run a single task foreground.",
            argument_default=UNDEFINED_ARGUMENT,
        )
        define_core_arguments(parser)
        parser.add_argument(
            'worker_name',
            metavar='WORKER',
            help=(
                "Global name of the worker function name to execute."
                " Use ? to list available workers."),
        )
        parser.add_argument(
            'worker_args', nargs='*',
            metavar='ARG',
            default=[],
            help="Worker arguments as Python literals.")

        args = parser.parse_args(argv)
        self.bootstrap(args=args, environ=environ)
        self.log_versions()
        self.apply_config()

        # Bootstraping plugins
        self.bootstrap_plugins()
        # Boostraping action logs table
        NotificationMgmt.bootstrap(self.config)

        # E X E C U T I O N

        workers = iter_workers(self.worker_pool.worker_pool.workers)

        if '?' == args.worker_name:
            for name in sorted(fn.__name__ for fn in workers):
                print(name)
            return 0

        needles = (args.worker_name, args.worker_name + '_worker')
        for worker in workers:
            if worker.__name__ in needles:
                break
        else:
            raise UserError("Unknown worker %s." % args.worker_name)

        worker_args = []
        for arg in args.worker_args:
            try:
                arg = literal_eval(arg)
            except Exception:
                logger.debug("Unknown literal %s, using as raw string.", arg)
            worker_args.append(arg)

        worker(self, *worker_args)

    def bootstrap_plugins(self):
        for plugin_name, plugin in self.plugins.items():
            if hasattr(plugin, 'bootstrap'):
                logger.debug("Boostraping plugin %s", plugin_name)
                plugin.bootstrap()


def iter_workers(workers):
    for name, config in workers.items():
        wrapper = getattr(sys.modules[config['module']], config['function'])
        yield wrapper._tm_function


main = TaskApplication(specs=list_options_specs())


if __name__ == "__main__":
    sys.exit(main())



import logging
import time

from ...toolkit import taskmanager
from ...toolkit.configuration import OptionSpec
from ...routing import RouteSet

from . import db
from . import metrics

logger = logging.getLogger(__name__)
routes = RouteSet(prefix=b'/dashboard')
workers = taskmanager.WorkerSet()


@routes.get(b'', check_key=True)
def dashboard(http_context, app):
    return metrics.get_metrics_queue(app.config)


@routes.get(b'/config', check_key=True)
def dashboard_config(http_context, app):
    return dict(
        scheduler_interval=app.config.dashboard.scheduler_interval,
        history_length=app.config.dashboard.history_length,
    )


@routes.get(b'/live', check_key=True)
def dashboard_live(http_context, app):
    return metrics.get_metrics(app)


@routes.get(b'/history', check_key=True)
def dashboard_history(http_context, app):
    return metrics.get_history_metrics_queue(app.config)


@routes.get(b'/buffers', check_key=True)
def dashboard_buffers(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_buffers(conn)


@routes.get(b'/hitratio', check_key=True)
def dashboard_hitratio(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_hitratio(conn)


@routes.get(b'/active_backends', check_key=True)
def dashboard_active_backends(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_active_backends(conn)


@routes.get(b'/cpu', check_key=True)
def dashboard_cpu(http_context, app):
    return metrics.get_cpu_usage()


@routes.get(b'/loadaverage', check_key=True)
def dashboard_loadaverage(http_context, app):
    return metrics.get_loadaverage()


@routes.get(b'/memory', check_key=True)
def dashboard_memory(http_context, app):
    return metrics.get_memory_usage()


@routes.get(b'/hostname', check_key=True)
def dashboard_hostname(http_context, app):
    return metrics.get_hostname(app.config)


@routes.get(b'/os_version', check_key=True)
def dashboard_os_version(http_context, app):
    return metrics.get_os_version()


@routes.get(b'/pg_version', check_key=True)
def dashboard_pg_version(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_pg_version(conn)


@routes.get(b'/n_cpu', check_key=True)
def dashboard_n_cpu(http_context, app):
    return metrics.get_n_cpu()


@routes.get(b'/databases', check_key=True)
def dashboard_databases(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_databases(conn)


@routes.get(b'/info', check_key=True)
def dashboard_info(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_info(conn, app.config)


@routes.get(b'/max_connections', check_key=True)
def dashboard_max_connections(http_context, app):
    with app.postgres.connect() as conn:
        return metrics.get_max_connections(conn)


@workers.register(pool_size=1)
def dashboard_collector_worker(app):
    logger.debug("Starting dashboard collector")

    data = metrics.get_metrics(app)

    # We don't want to store notifications in the history.
    data.pop('notifications', None)
    logger.debug(data)

    db.add_metric(
        app.config.temboard.home,
        'dashboard.db',
        time.time(),
        data,
        app.config.dashboard.history_length
    )

    logger.debug("Done")


class DashboardPlugin:
    PG_MIN_VERSION = (90400, 9.4)
    s = 'dashboard'
    option_specs = [
        OptionSpec(s, 'scheduler_interval', default=2, validator=int),
        OptionSpec(s, 'history_length', default=150, validator=int),
    ]
    del s

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.option_specs)

    def bootstrap(self):
        db.bootstrap(self.app.config.temboard.home, 'dashboard.db')

    def load(self):
        self.app.router.add(routes)
        self.app.worker_pool.add(workers)
        workers.schedule(
            id='dashboard_collector',
            redo_interval=self.app.config.dashboard.scheduler_interval
        )(dashboard_collector_worker)
        self.app.scheduler.add(workers)

    def unload(self):
        self.app.scheduler.remove(workers)
        self.app.worker_pool.remove(workers)
        self.app.router.remove(routes)
        self.app.config.remove_specs(self.option_specs)



# coding: utf-8

from .pycompat import IterableUserDict


_UNDEFINED = object()


def dict_factory(iterable=_UNDEFINED, **kw):
    # a dict factory that do not copy.
    if iterable is _UNDEFINED:
        return kw
    elif isinstance(iterable, dict):
        return iterable
    else:
        return dict(iterable, **kw)


class DotDict(IterableUserDict):
    # A wrapper around dict that allows read and write through dot style
    # accessors.

    def __init__(self, *a, **kw):
        self.__dict__['data'] = dict_factory(*a, **kw)

    def __getattr__(self, name):
        try:
            value = self[name]
            # Lazy recursion of DotDict
            if isinstance(value, dict):
                self[name] = value = DotDict(value)
            return value
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if hasattr(value, 'items'):
            value = DotDict(value)
        self[name] = value

    def __setstate__(self, state):
        self.__dict__.update(state)

    def setdefault(self, name, default):
        if hasattr(default, 'items'):
            default = DotDict(default)
        return IterableUserDict.setdefault(self, name, default)



from ...routing import RouteSet

from . import functions as activity_functions


routes = RouteSet()


@routes.get(b'/activity', check_key=True)
def get_activity(http_context, app):
    with app.postgres.connect() as conn:
        return activity_functions.get_activity(conn)


@routes.get(b'/activity/waiting', check_key=True)
def get_activity_waiting(http_context, app):
    with app.postgres.connect() as conn:
        return activity_functions.get_activity_waiting(conn)


@routes.get(b'/activity/blocking', check_key=True)
def get_activity_blocking(http_context, app):
    with app.postgres.connect() as conn:
        return activity_functions.get_activity_blocking(conn)


@routes.post(b'/activity/kill')
def post_activity_kill(http_context, app):
    with app.postgres.connect() as conn:
        return activity_functions.post_activity_kill(conn, app.config,
                                                     http_context)


class ActivityPlugin:
    PG_MIN_VERSION = (90400, 9.4)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        self.app.router.add(routes)

    def unload(self):
        self.app.router.remove(routes)



import errno
import functools
import sys
import time
import uuid
import logging
import os.path
import types
import signal
from select import select, error as SelectError
from datetime import datetime
from collections import deque
from multiprocessing import AuthenticationError, Process, Queue
from multiprocessing.connection import Listener, Client

from .services import Service
from .errors import StorageEngineError
from .perf import PerfCounters
from .pycompat import PY2, Empty


TM_WORKERS = []
TM_BOOTSTRAP = []
TM_SIGHUP = False
TM_SIGTERM = False
TM_SIGABRT = False

TM_DEF_LISTENER_ADDR = '/tmp/.temboardsched.sock'

# Message types
MSG_TYPE_TASK_NEW = 0
MSG_TYPE_TASK_STATUS = 1
MSG_TYPE_TASK_CANCEL = 2
MSG_TYPE_TASK_ABORT = 3
MSG_TYPE_TASK_LIST = 4
MSG_TYPE_RESP = 5
MSG_TYPE_ERROR = 6
MSG_TYPE_CONTEXT = 7

# Task status
TASK_STATUS_DEFAULT = 1
TASK_STATUS_SCHEDULED = 2
TASK_STATUS_QUEUED = 4
TASK_STATUS_DOING = 8
TASK_STATUS_DONE = 16
TASK_STATUS_FAILED = 32
TASK_STATUS_CANCELED = 64
TASK_STATUS_ABORTED = 128
TASK_STATUS_ABORT = 256

logger = logging.getLogger(__name__)


def ensure_str(value):
    # This code is used to instanciate multiprocessing.connection.Client. It
    # requires a str object in both Python 2 and 3.
    if type(value) is not str:
        if PY2:
            # From unicode to str.
            value = value.encode('utf-8')
        else:
            # From bytes to str.
            value = value.decode('utf-8')
    return value


def json_serial_datetime(obj):
    # JSON serializer for datetime
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def make_worker_definition(function, pool_size):
    return {
        'name': function.__name__,
        'pool_size': pool_size,
        'module': function.__module__,
        'function': function.__name__
    }


def worker(pool_size=1):
    # Decorator that defines a new worker function
    def defines_worker(function):
        global TM_WORKERS
        TM_WORKERS.append(make_worker_definition(function, pool_size))
        return function

    return defines_worker


def handler_sigterm(num, frame):
    global TM_SIGTERM
    TM_SIGTERM = True


def handler_sighup(num, frame):
    global TM_SIGHUP
    TM_SIGHUP = True


def handler_sigabrt(num, frame):
    global TM_SIGABRT
    TM_SIGABRT = True


def bootstrap():
    def defines_bootstrap(function):
        global TM_BOOTSTRAP
        TM_BOOTSTRAP.append({
            'module': function.__module__,
            'function': function.__name__
        })
        return function

    return defines_bootstrap


def schedule_task(worker_name, id=None, options=None, start=None,
                  redo_interval=None, listener_addr=TM_DEF_LISTENER_ADDR,
                  authkey=None, expire=3600):
    # Schedule a new task
    return TaskManager.send_message(
                listener_addr,
                Message(
                    MSG_TYPE_TASK_NEW,
                    Task(
                        id=id,
                        worker_name=worker_name,
                        options=options,
                        start_datetime=start or datetime.utcnow(),
                        redo_interval=redo_interval,
                        expire=expire,
                    )
                ),
                authkey=authkey
           )


def set_context(k, v, listener_addr=TM_DEF_LISTENER_ADDR, authkey=None):
    # Update a context variable
    return TaskManager.send_message(
                listener_addr,
                Message(
                    MSG_TYPE_CONTEXT,
                    {k: v}
                ),
                authkey=authkey
           )


class Task(object):

    def __init__(self, worker_name=None, options=None, id=None,
                 status=TASK_STATUS_DEFAULT, start_datetime=None,
                 redo_interval=None, stop_datetime=None, output=None,
                 expire=3600):
        self.worker_name = worker_name
        self.options = options
        self.status = status
        self.start_datetime = start_datetime or datetime.utcnow()
        self.redo_interval = redo_interval
        self.stop_datetime = stop_datetime
        self.id = id
        self.output = output
        # Task expiration timeout in seconds.
        # Ended Tasks are removed from memory when current time exceeds
        # self.stop_datetime + self.expire.
        self.expire = expire

    def __repr__(self):
        return str(self.__dict__)


class Message(object):

    def __init__(self, type, content):
        self.type = type,
        self.content = content

    def __repr__(self):
        return str(self.__dict__)


class TaskList(object):

    def __init__(self, engine):
        self.engine = engine

    def recover(self):
        self.engine.recover(
            st_doing=(TASK_STATUS_DOING | TASK_STATUS_QUEUED),
            st_aborted=TASK_STATUS_ABORTED,
            st_scheduled=TASK_STATUS_SCHEDULED,
            st_default=TASK_STATUS_DEFAULT,
            now=datetime.utcnow()
        )

    def push(self, task):
        # Add a new task to the list
        if not task.id:
            task.id = self._gen_task_id()
        # Insert the task
        self.engine.insert(task)
        return task.id

    def get(self, id):
        return self.engine.get(id)

    def update(self, id, **kwargs):
        task = self.engine.get(id)
        if not task:
            raise Exception("Task id=%s not found" % id)

        for k, v in kwargs.items():
            try:
                getattr(task, k)
            except AttributeError:
                raise Exception("Task attribute %s does not exist" % k)
            setattr(task, k, v)
        self.engine.update(task)

    def rm(self, id):
        self.engine.delete(id)

    def _gen_task_id(self):
        id = str(uuid.uuid4())[0:8]

        while self.engine.exists(id):
            # If a task with the same id exists, try with a new random id
            id = str(uuid.uuid4())[0:8]

        return id

    def get_n_todo(self):
        # Return the number of ongoing tasks (QUEUED | DOING)
        return self.engine.count_by_status(
            TASK_STATUS_QUEUED | TASK_STATUS_DOING
        )

    def list(self):
        return self.engine.list()

    def list_to_do(self, status, now, redo=False):
        return self.engine.list_to_do(status, now, redo)

    def purge(self, status, now):
        return self.engine.purge(status, now)


class TaskManager(object):

    def __init__(self, address=TM_DEF_LISTENER_ADDR, authkey=None):

        self.scheduler = Scheduler(address, authkey)

    def set_context(self, key, val):
        self.scheduler.set_context(key, val)

    def start(self):
        logger.info("Starting TaskManager")
        self.scheduler.start()

    def restart(self, abort=False):
        if abort:
            self.stop()
            self.start()
        else:
            self.shutdown()
            self.start()

    def shutdown(self):
        # Will stop the scheduler and WP only when all the ongoing task are
        # done.
        logger.info("Shutting down scheduler")
        os.kill(self.scheduler.process.pid, signal.SIGTERM)
        self.join()

    def stop(self):
        # Abort all ongoing tasks and stop the scheduler and WP
        logger.info("Aborting scheduler")
        os.kill(self.scheduler.process.pid, signal.SIGABRT)
        self.join()

    def join(self):
        try:
            logger.debug("Joining processes")
            self.scheduler.process.join()
            self.scheduler.worker_pool.process.join()
        except KeyboardInterrupt:
            logger.debug("KeyboardInterrupt")
            return

    @staticmethod
    def send_message(address, message, authkey=''):
        conn = Client(ensure_str(address), authkey=authkey)
        conn.send(message)
        res = conn.recv()
        conn.close()
        return res


class Scheduler(object):

    def __init__(self, address, authkey):
        # Process objet from multiprocessing
        self.process = None
        # Listener for TM -> Scheduler IPC
        self.listener = None
        # Listener address
        self.address = address
        # Listener authentication key
        self.authkey = authkey
        # Queue used to send Task orders (start, stop) to WorkerPool
        self.task_queue = None
        # Queue used to notify Scheduler about Task status
        self.event_queue = None
        self.task_list = None
        self.context = dict()
        self.worker_pool = None
        self.shutdown = None
        self.task_list_engine = None
        self.last_schedule = 0
        self.last_vacuum = 0
        self.select_timeout = None

    def set_context(self, key, val):
        self.context[key] = val

    def get_context(self):
        return self.context

    def handle_listener_message(self):
        # read message from the listener
        try:
            conn = self.listener.accept()
        except AuthenticationError as e:
            logger.exception(e)
            logger.error("Authentication failed")
            return

        try:
            message = conn.recv()
        except Exception as e:
            conn.close()
            logger.exception(e)
            logger.error("Unable to read the message")
            return

        message.type = message.type[0]
        logger.debug("Received Message=%s" % message)

        # handle incoming message and return a response
        conn.send(self.handle_message(message))
        conn.close()

    def handle_event_queue_message(self):
        # read message from the event queue
        try:
            message = self.event_queue.get(False)
        except Empty:
            # should not happen
            logger.error("Event queue empty")
            return

        message.type = message.type[0]
        logger.debug("Received Message=%s" % message)

        # handle incoming message
        self.handle_message(message)

    def sync_bootstrap_options(self):
        # Reload bootstrap Task options and update Task from task_list with new
        # options. This is usefull to reflect context changes into Task options
        # like in a configuration update case.
        for bs_func in TM_BOOTSTRAP:
            func = getattr(sys.modules[bs_func['module']],
                           bs_func['function'])(context=self.get_context())
            if isinstance(func, types.GeneratorType):
                for t in func:
                    if isinstance(t, Task):
                        try:
                            self.task_list.update(
                                    t.id,
                                    options=t.options,
                                    redo_interval=t.redo_interval,
                            )
                        except Exception as e:
                            logger.exception(str(e))
                            logger.error("Could not update Task %s with"
                                         " options=%s" % (t.id, t.options))
                    else:
                        logger.error("Bootstrap task not Task instance")
            else:
                logger.error("Bootstrap function %s.%s not a generator"
                             % (bs_func['module'], bs_func['function']))

    def bootstrap(self):
        # Load Tasks from generators decorated by @taskmanager.bootstrap()
        for bs_func in TM_BOOTSTRAP:
            func = getattr(sys.modules[bs_func['module']],
                           bs_func['function'])(context=self.get_context())
            if isinstance(func, types.GeneratorType):
                for t in func:
                    if isinstance(t, Task):
                        try:
                            try:
                                self.task_list.push(t)
                                logger.debug("Bootstrap Task=%s loaded" % t)
                            except KeyError:
                                logger.debug("Update Task %s with options=%s"
                                             % (t.id, t.options))
                                self.task_list.update(
                                    t.id,
                                    options=t.options,
                                    redo_interval=t.redo_interval,
                                )
                        except StorageEngineError as e:
                            # Just log the error and continue with other tasks
                            logger.error(str(e))
                    else:
                        logger.error("Bootstrap task not Task instance")
            else:
                logger.error("Bootstrap function %s.%s not a generator"
                             % (bs_func['module'], bs_func['function']))

    def setup(self):
        # Need to shutdown ?
        self.shutdown = False
        # Instanciate a new Listener
        self.listener = Listener(self.address, family='AF_UNIX',
                                 authkey=self.authkey)
        # bootstrap
        self.bootstrap()
        # TODO
        # self.sync_bootstrap_options()
        self.select_timeout = 1

    def serve1(self):
        # wait for I/O on Listener and event Queue
        try:
            fds, _, _ = select(
                [self.listener._listener._socket.fileno(),
                 self.event_queue._reader.fileno()],
                [],
                [],
                self.select_timeout
            )
        except SelectError as e:
            errno_, message = e.args
            if errno_ == errno.EINTR:
                # Interrupted by e.g. SIGHUP. Just stop.
                return
            else:
                raise

        if len(fds) > 0:
            for fd in fds:
                if fd == self.listener._listener._socket.fileno():
                    if self.shutdown:
                        # during shutdown we don't handle incoming
                        # message anymore
                        continue
                    self.handle_listener_message()
                elif fd == self.event_queue._reader.fileno():
                    self.handle_event_queue_message()

        if self.last_schedule < (time.time() - 1):
            self.schedule()
            self.last_schedule = time.time()

        if self.last_vacuum < (time.time() - 3600):
            self.task_list.engine.vacuum()
            self.last_vacuum = time.time()

    def setup_task_list(self):
        # Instanciate TaskList
        self.task_list = TaskList(self.task_list_engine)
        # Bootstrap storage
        self.task_list.engine.bootstrap()
        # Reset task status
        self.task_list.recover()

    def run(self):
        # Add signal handlers
        signal.signal(signal.SIGTERM, handler_sigterm)
        signal.signal(signal.SIGHUP, handler_sighup)
        signal.signal(signal.SIGABRT, handler_sigabrt)

        global TM_SIGHUP
        global TM_SIGTERM
        global TM_SIGABRT

        self.setup_task_list()
        self.setup()

        while True:
            try:
                self.serve1()

                if TM_SIGTERM:
                    # We perform shutdown on SIGTERM: we wait until all
                    # ongoing tasks are done before stopping
                    logger.debug("SIGTERM received")
                    self.shutdown = True
                    TM_SIGTERM = False
                if TM_SIGHUP:
                    logger.debug("SIGHUP received")
                    self.sync_bootstrap_options()
                    TM_SIGHUP = False

                if TM_SIGABRT:
                    logger.debug("SIGABRT received")
                    TM_SIGABRT = False
                    self.stop()
                    return

                try:
                    # Check parent
                    os.kill(os.getppid(), 0)
                except Exception:
                    self.stop()
                    return

                if self.shutdown and self.task_list.get_n_todo() == 0:
                    # When shutting down, if there is no more ongoing task
                    # then we can stop
                    self.stop()
                    return
            except KeyboardInterrupt:
                logger.error("KeyboardInterrupt")
                self.stop()
                return
            except Exception as e:
                logger.exception(e)
                self.stop()
                return

    def schedule(self):
        now = datetime.utcnow()

        self.task_list.purge(
            (TASK_STATUS_DONE | TASK_STATUS_FAILED | TASK_STATUS_ABORTED |
             TASK_STATUS_CANCELED),
            now
        )

        if self.shutdown:
            return

        to_do = self.task_list.list_to_do(TASK_STATUS_DEFAULT, now)

        for task in to_do:
            task.status = TASK_STATUS_SCHEDULED

            logger.debug("Pushing task to the worker queue")
            logger.debug(task)

            try:
                self.task_list.update(
                    id=task.id,
                    status=task.status
                )
            except StorageEngineError as e:
                logger.error(str(e))
            else:
                self.task_queue.put(task, False)

        to_redo = self.task_list.list_to_do(
            (TASK_STATUS_DONE | TASK_STATUS_FAILED | TASK_STATUS_ABORTED),
            now,
            redo=True
        )

        for task in to_redo:
            task.status = TASK_STATUS_SCHEDULED
            task.start_datetime = now
            task.stop_datetime = None
            task.output = ''

            logger.debug("Pushing task to the worker queue")
            logger.debug(task)

            try:
                self.task_list.update(
                    id=task.id,
                    status=task.status,
                    start_datetime=task.start_datetime,
                    stop_datetime=task.stop_datetime,
                    output=task.output,
                )
            except StorageEngineError as e:
                logger.error(str(e))
            else:
                self.task_queue.put(task, False)

    def handle_message(self, message):
        if message.type == MSG_TYPE_TASK_NEW:
            # New task
            try:
                task_id = self.task_list.push(message.content)
                return Message(MSG_TYPE_RESP, {'id': task_id})
            except KeyError:
                return Message(MSG_TYPE_ERROR,
                               {'error': 'Task id already exists'})
            except StorageEngineError as e:
                logger.error(str(e))
                return Message(MSG_TYPE_ERROR, {'error': str(e)})

        elif message.type == MSG_TYPE_TASK_STATUS:
            # task status update
            status = message.content['status']
            # special case when task's status is TASK_STATUS_CANCELD, we dont'
            # want to change it's state.
            try:
                t = self.task_list.get(message.content['task_id'])
                if t.status & TASK_STATUS_CANCELED:
                    status = t.status

                self.task_list.update(
                    t.id,
                    status=status,
                    output=message.content.get('output', None),
                    stop_datetime=message.content.get('stop_datetime', None),
                )
            except StorageEngineError as e:
                logger.error(str(e))
                return Message(MSG_TYPE_ERROR, {'error': str(e)})
            else:
                return Message(MSG_TYPE_RESP, {'id': t.id})

        elif message.type == MSG_TYPE_TASK_LIST:
            # task list
            try:
                return list(self.task_list.list())
            except StorageEngineError as e:
                logger.error(str(e))
                return Message(MSG_TYPE_ERROR, {'error': str(e)})

        elif message.type == MSG_TYPE_TASK_ABORT:
            # task abortation
            t = Task(id=message.content['task_id'],
                     status=TASK_STATUS_ABORT)
            self.task_queue.put(t)

        elif message.type == MSG_TYPE_TASK_CANCEL:
            # task cancellation
            # first, we need to change its status and stop_datetime
            try:
                self.task_list.update(
                    message.content['task_id'],
                    status=TASK_STATUS_CANCELED,
                    stop_datetime=datetime.utcnow(),
                )
            except StorageEngineError as e:
                logger.error(str(e))
            else:
                # send the cancelation order to WP
                t = Task(id=message.content['task_id'],
                         status=TASK_STATUS_CANCELED)
                self.task_queue.put(t)
        elif message.type == MSG_TYPE_CONTEXT:
            # context update
            if type(message.content) == dict:
                for k, v in message.content.items():
                    self.set_context(k, v)
                return Message(MSG_TYPE_RESP, self.get_context())
            else:
                return Message(MSG_TYPE_ERROR, 'Unvalid type')

        # TODO: handle other message types

    def start(self):
        # Queues
        self.task_queue = Queue()
        self.event_queue = Queue()
        # Instanciate WP
        self.worker_pool = WorkerPool(self.task_queue, self.event_queue)
        # Start WP as new process
        self.worker_pool.start()
        # Start Scheduler as new Process
        self.process = Process(target=self.run, args=())
        self.process.start()

    def stop(self):
        logger.debug("Abort remaining jobs if any")
        self.worker_pool.abort_jobs()
        logger.debug("Terminating WP process")
        self.worker_pool.process.terminate()
        logger.debug("Closing Listener")
        self.listener.close()
        logger.debug("Closing task_queue")
        self.task_queue.close()
        logger.debug("Closing event_queue")
        self.event_queue.close()
        del self.listener
        del self.worker_pool
        del self.task_queue
        del self.event_queue
        del self.task_list


class SchedulerService(Service):
    # Adapter from taskmanager.Scheduler to Service

    def __init__(self, task_queue, event_queue, **kw):
        super(SchedulerService, self).__init__(**kw)
        self.task_queue = task_queue
        self.event_queue = event_queue
        self.scheduler = None
        self.task_list_engine = None

    def apply_config(self):
        # Setup scheduler as soon as configuration is loaded, before
        # plugins, so that tasklist is created before plugins.
        if not self.scheduler:
            self.scheduler = Scheduler(
                address=os.path.join(
                    self.app.config.temboard.home, '.tm.socket'),
                authkey=None)
            self.scheduler.task_queue = self.task_queue
            self.scheduler.event_queue = self.event_queue
            self.scheduler.task_list_engine = self.task_list_engine
            self.scheduler.setup_task_list()

    def setup(self):
        if os.path.exists(self.scheduler.address):
            os.unlink(self.scheduler.address)

        self.scheduler.setup()

    def serve1(self):
        self.scheduler.serve1()

    def add(self, workerset):
        if not self.is_my_process:
            return

        for task in workerset.list_tasks():
            try:
                task_from_db = self.scheduler.task_list.get(task.id)

                if task_from_db:
                    # If we've found the task in the DB, we'd keep its
                    # stop_datetime and status before overwriting it because
                    # we do not want this task to be re scheduled if it's not
                    # necessary.
                    task.stop_datetime = task_from_db.stop_datetime
                    task.status = task_from_db.status

                    self.scheduler.task_list.rm(task.id)
                    logger.debug("Overwriting task %s.", task.id)

                self.scheduler.task_list.push(task)
            except StorageEngineError as e:
                logger.error(str(e))

    def remove(self, workerset):
        if not self.is_my_process:
            return

        for task in workerset.list_tasks():
            try:
                self.scheduler.task_list.rm(task.id)
            except StorageEngineError as e:
                logger.error(str(e))

    def schedule_task(
            self, worker_name, id=None, options=None, start=None,
            redo_interval=None, expire=3600):
        message = Message(MSG_TYPE_TASK_NEW, Task(
            id=id,
            worker_name=worker_name,
            options=options,
            start_datetime=start or datetime.utcnow(),
            redo_interval=redo_interval,
            expire=expire,
        ))

        conn = Client(
            ensure_str(self.scheduler.address), self.scheduler.authkey)
        conn.send(message)
        res = conn.recv()
        conn.close()
        return res


class WorkerPool(object):

    def __init__(self, task_queue, event_queue, setproctitle=None):
        self.thread = None
        self.task_queue = task_queue
        self.event_queue = event_queue
        self.workers = {}
        self.setproctitle = setproctitle
        self.perf = None

    def _abort_job(self, task_id):
        for workername in self.workers:
            for job in self.workers[workername]['pool']:
                if job['id'] == task_id:
                    logger.debug("Process pid=%s is going to be killed"
                                 % job['process'])
                    job['process'].terminate()
                    return True
        return False

    def _rm_task_worker_queue(self, task_id):
        for workername in self.workers:
            for t in self.workers[workername]['queue']:
                if t.id == task_id:
                    self.workers[workername]['queue'].remove(t)
                    logger.debug("Task %s removed from queue" % t.id)
                    return True
        return False

    def setup(self):
        if self.perf:
            self.perf['fork'] = 0
        global TM_WORKERS
        for worker in TM_WORKERS:
            self.add(worker)

    def add(self, worker):
        self.workers[worker['name']] = {
            'queue': deque(),
            'pool_size': worker['pool_size'],
            'module': worker['module'],
            'function': worker['function'],
            'pool': []
        }

    def serve1(self):
        # check running jobs state
        self.check_jobs()
        # start new jobs
        self.start_jobs()

        try:
            t = self.task_queue.get(timeout=0.1)
        except Empty:
            return

        if t.status & TASK_STATUS_SCHEDULED:
            logger.debug("Add Task %s to worker '%s' queue"
                         % (t.id, t.worker_name))
            self.workers[t.worker_name]['queue'].appendleft(t)
            # Update task status
            self.event_queue.put(
                Message(
                    MSG_TYPE_TASK_STATUS,
                    {
                        'task_id': t.id,
                        'status': TASK_STATUS_QUEUED,
                    }
                )
            )
        if t.status & TASK_STATUS_ABORT:
            self._abort_job(t.id)
        if t.status & TASK_STATUS_CANCELED:
            # Task cancellation includes 2 things:
            # - remove the task from workers queue if present (job not
            # yet running)
            # - abort running job executing the task if any

            # Looking up into workers jobs first, there is more chance
            # to find the task here.
            if not self._abort_job(t.id):
                # If not aborted, task has been queued
                self._rm_task_worker_queue(t.id)

    def run(self):
        self.setup()
        while True:
            try:
                self.serve1()
            except Exception as e:
                logger.exception(e)
            except KeyboardInterrupt:
                logger.error("KeyboardInterrupt")
                return

    def exec_worker(self, module, function, out, *args, **kws):
        # Function wrapper around worker function
        try:
            fun = getattr(sys.modules[module], function)

            modfun = "%s.%s" % (module, fun.__name__)
            logger.debug("Starting new worker %s", modfun)
            if self.setproctitle:
                self.setproctitle('task %s' % (modfun))
            perf = PerfCounters.setup(service='task', task=modfun)
            if perf:
                perf.schedule()

            res = fun(*args, **kws)
            # Put function result into output queue as a Message
            out.put(Message(MSG_TYPE_RESP, res))
        except Exception as e:
            e = Exception("%s: %s" % (type(e), e))
            out.put(Message(MSG_TYPE_ERROR, e))
            logger.exception(e)
        except KeyboardInterrupt:
            logger.error("KeyboardInterrupt")
        if perf:
            perf.run()

    def start_jobs(self):
        # Execute Tasks
        for name, worker in self.workers.items():
            while len(self.workers[name]['pool']) < worker['pool_size']:
                try:
                    t = worker['queue'].pop()
                    # Queue used to get worker function return
                    out = Queue()
                    p = Process(
                            target=self.exec_worker,
                            args=(worker['module'], worker['function'], out),
                            kwargs=t.options,
                        )
                    p.start()

                    if self.perf:
                        self.perf['fork'] += 1

                    self.workers[name]['pool'].append(
                            {'id': t.id, 'process': p, 'out': out}
                    )
                    # Update task status
                    self.event_queue.put(
                        Message(
                            MSG_TYPE_TASK_STATUS,
                            {
                                'task_id': t.id,
                                'status': TASK_STATUS_DOING,
                            }
                        )
                    )
                except IndexError:
                    break

    def check_jobs(self):
        # Check jobs process state for each worker
        for name, worker in self.workers.items():
            for job in worker['pool']:
                if not job['process'].is_alive():
                    # Dead process case
                    logger.debug("Job %s is dead" % job['id'])
                    try:
                        # Fetch the message from job's output queue
                        message_out = job['out'].get(False)
                    except Empty:
                        message_out = None
                    logger.debug("Job output : %s" % message_out)
                    # Close job's output queue
                    job['out'].close()
                    # join the process
                    job['process'].join()

                    # Let's build the message we'll have to send to scheduler
                    # for the update of task's status.
                    task_stop_dt = datetime.utcnow()
                    if job['process'].exitcode == 0:
                        if message_out and \
                                message_out.type[0] == MSG_TYPE_RESP:
                            task_status = TASK_STATUS_DONE
                        else:
                            # when an exception is raised from the worker
                            # function
                            task_status = TASK_STATUS_FAILED
                    elif job['process'].exitcode < 0:
                        # process killed
                        task_status = TASK_STATUS_ABORTED
                    else:
                        task_status = TASK_STATUS_FAILED
                    task_output = None
                    if message_out:
                        task_output = message_out.content

                    # Update task status
                    self.event_queue.put(
                        Message(
                            MSG_TYPE_TASK_STATUS,
                            {
                                'task_id': job['id'],
                                'status': task_status,
                                'output': task_output,
                                'stop_datetime': task_stop_dt
                            }
                        )
                    )

                    # Finally, remove the job from the pool
                    self.workers[name]['pool'].remove(job)

    def start(self):
        self.process = Process(target=self.run, args=())
        self.process.start()

    def abort_jobs(self):
        # abort all running jobs
        for name, worker in self.workers.items():
            for job in worker['pool']:
                process = job.get('process')
                if process.is_alive():
                    process.terminate()
                    logger.debug("Job %s has been terminated" % job['id'])
                    # Close job's output queue
                    job['out'].close()
                    # join the process
                    process.join()
                    self.workers[name]['pool'].remove(job)


class WorkerSet(list):
    def register(self, pool_size=1):
        def register(f):
            f._tm_worker = make_worker_definition(f, pool_size)
            if f not in self:
                self.append(f)
            return f
        return register

    def schedule(self, id=None, redo_interval=None, **options):
        def register(f):
            if f not in self:
                self.append(f)
            f._tm_task = Task(
                id=id,
                options=options,
                redo_interval=redo_interval,
                worker_name=f.__name__,
            )
            return f
        return register

    def list_tasks(self):
        for function in self:
            task = getattr(function, '_tm_task', None)
            if task:
                yield task


class WorkerPoolService(Service):
    # Adapter from taskmanager.WorkerPool to Service

    def __init__(self, task_queue, event_queue, **kw):
        super(WorkerPoolService, self).__init__(**kw)
        self.worker_pool = WorkerPool(
            task_queue, event_queue, self.setproctitle)

    def setup(self):
        self.worker_pool.perf = self.perf
        self.worker_pool.setup()

    def serve1(self):
        try:
            self.worker_pool.serve1()
        except Exception:
            logger.exception("Unhandled error in worker:")
            logger.error("Not stopping worker process.")

    def create_task_function_app_wrapper(self, function):
        @functools.wraps(function)
        def wrapper(*a, **kw):
            return function(app=self.app, *a, **kw)
        wrapper._tm_function = function
        return wrapper

    def add(self, workerset):
        if not self.is_my_process:
            return

        for function in workerset:
            conf = function._tm_worker
            wrapper = self.create_task_function_app_wrapper(function)

            # Inject wrapper in module so taskmanager will find it.
            mod = sys.modules[conf['module']]
            wrapper_name = '_tm_wrapper_' + function.__name__
            setattr(mod, wrapper_name, wrapper)
            conf['function'] = wrapper_name

            # Add to current workers
            logger.debug("Activate worker %s", conf['name'])
            self.worker_pool.add(conf)

    def remove(self, workerset):
        if not self.is_my_process:
            return

        for function in workerset:
            conf = function._tm_worker
            logger.debug("Disable worker %s", conf['name'])
            self.worker_pool.workers.pop(conf['name'], None)



import errno
import hashlib
import json
import logging
import os
import re
import time
from datetime import datetime
from time import strftime, gmtime
from .errors import HTTPError

logger = logging.getLogger(__name__)


def hash_id(id):
    """
    Hash (MD5) and returns a string built from the given id aimed to be used as
    an ID.
    """
    strid = 'id%s%d' % (id, time.time() * 1000)
    m = hashlib.md5()
    m.update(strid.encode('utf-8'))
    return m.hexdigest()


def validate_parameters(values, types):
    """
    Verify that each value of dict 'values' is valid. For doing that, we have
    to loop over all 'types' elements which are tuples: ('values' key,
    validation regexp, if the value item currently checked is a list of thing
    to check).
    If values[key] (or each element of it when it's a list) does not match
    with the regexp then we trow an error.
    """
    for (key, typ, is_list) in types:
        try:
            if type(typ) == bytes and hasattr(typ, 'decode'):
                typ = str(typ.decode('utf-8'))
            if not is_list:
                # If 'typ' is a string, it must be considered as a regexp
                # pattern.
                if type(typ) == str and \
                        re.match(typ, str(values[key])) is None:
                    raise HTTPError(406, "Parameter '%s' is malformed."
                                         % (key))
                if type(typ) != str and isinstance(values[key], type(typ)):
                    raise HTTPError(406, "Parameter '%s' is malformed."
                                         % (key))
            if is_list:
                for value in values[key]:
                    if type(typ) == str and re.match(typ, str(value)) is None:
                        raise HTTPError(406, "Parameter '%s' is malformed."
                                             % (key))
                    if type(typ) != str and typ != type(value):
                        raise HTTPError(406, "Parameter '%s' is malformed."
                                             % (key))
        except HTTPError as e:
            raise e
        except KeyError:
            raise HTTPError(406, "Parameter '%s' not sent." % (key))
        except Exception as e:
            logger.exception(str(e))
            raise HTTPError(500, "Internal error.")


MULTIPLIERS = ['', 'k', 'M', 'G', 'T', 'P']


def to_bytes(size, unit):
    """
    Convert the given size, expressed in unit, to bytes.

    Args:
        size (int): the size to convert to bytes
        unit (str): the unit this size is in
    Returns:
        the converted size

    Example:

    >>> to_bytes(7890, 'M')
    8273264640
    >>> to_bytes(7890, 'unexistent_unit')
    Traceback (most recent call last):
    ...
    KeyError: 'Invalid unit: unexistent_unit'

    """
    if unit not in MULTIPLIERS:
        raise KeyError("Invalid unit: %s" % unit)
    return size * 1024 ** MULTIPLIERS.index(unit)


def which(prog, search_path=None):
    """
    Search for a file in the given list of directories or fallback to PATH.

    Args:
        prog (str): the file to search
        search_path (list): list of directories to search
    Returns:
        the first path found

    Example:

    >>> which('ls') #doctest: +SKIP
    '/bin/ls'
    >>> which('ifconfig') #doctest: +SKIP
    Traceback (most recent call last):
    ...
    OSError: No such file or directory
    >>> which('ifconfig', ['/sbin']) #doctest: +SKIP
    '/sbin/ifconfig'
    """

    env_path = re.split(r':', os.environ['PATH'])
    if search_path is None:
        search_path = env_path
    else:
        search_path += env_path

    for d in search_path:
        path = d + '/' + prog
        if os.path.exists(path):
            return path

    raise OSError(os.strerror(errno.ENOENT))


def check_fqdn(name):
    """
    Check if a hostname is fully qualified, it must only contain
    letters, - and have dots.
    """
    # StackOverflow #11809631
    if re.match(r'(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]'  # noqa W605
                r'{2,63}\.?$)', name):
        return True
    else:
        return False


def now():
    """Give the current date and time at GMT, suitable for PostgreSQL."""
    return strftime("%Y-%m-%d %H:%M:%S +0000", gmtime())


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        else:
            return super().default(obj)



# coding: utf-8
#
# This is a pure python implementation of process title edition
#
# It is based on https://github.com/dvarrazzo/py-setproctitle which contains
# codes from Postgres itself. Postgres as a good experience one editing process
# title on various plateform. However, it's in C.
#
# There is not a single portable way accross OS, Python version and even ps
# implementation.
#
# The simplest and most portable implemention is argv overwriting. Each process
# is initialized with a memory chunk containing, in order : argv, argc,
# environ. This is what appears in /proc/self/cmdline and /proc/self/environ.
#
# Finding the actual address and size of argv is tricky. Once the memory chunk
# is located, we can just write process title in it and pad it with \0.

from __future__ import print_function

import ctypes
import logging
import sys
import os

if __package__:
    from .pycompat import PY3
else:
    PY3 = sys.version_info > (3,)


libc = ctypes.CDLL('libc.so.6')
logger = logging.getLogger(__name__)


def compute_main_module_name(mod):
    # Fix __main__ module to find its importable name.

    if mod.__name__ == '__main__':
        dir_, file_ = mod.__file__.rsplit('/', 1)
        name = file_.replace('.py', '')
    else:
        name = mod.__name__

    return mod.__package__ + '.' + name if mod.__package__ else name


def fix_argv(argv):
    # CPython alters argv in sys.argv. This breaks searching for sys.argv in
    # memory. This function reverses CPython alteration to libc original argv.
    #
    # Alterations are: adding a -c argument, trimming of main module name.

    modname = None
    # Whether the current -c is placeholder for command string
    command_string = False
    for i, arg in enumerate(argv[:]):
        if 0 == i:
            # Python interpreter. Skip it.
            continue
        elif not arg.startswith('-'):
            # Python argument. Next argv items are scripts arguments. Stop now.
            break
        elif '-' == arg:
            break
        elif '-c' == arg:
            if command_string:
                argv[i] = '__COMMAND_STRING__'
                break
            else:
                command_string = True
        elif '-m' == arg and modname is None:
            # Restore main module name.
            modname = compute_main_module_name(sys.modules['__main__'])
            # In PY3, -m module is replaced with -m -m.
            argv[i + 1] = modname
            break

    return argv


def find_argv_memory_from_pythonapi():  # pragma: nocover
    """ Return pointer and size of argv memory segment. """
    # This implemententation works only on Python2. cf.
    # http://docs.cherrypy.org/en/latest/_modules/cherrypy/process/wspbus.html

    # Allocate variable to point to argv
    argv_type = ctypes.c_wchar_p if PY3 else ctypes.c_char_p
    argv = ctypes.POINTER(argv_type)()
    argc = ctypes.c_int()

    # Get them from CPython API.
    ctypes.pythonapi.Py_GetArgcArgv(ctypes.byref(argc), ctypes.byref(argv))

    argl = [argv[i] for i in range(argc.value)]
    fix_argv(argl)

    # Point to first entry of argv.
    address = ctypes.addressof(argv.contents)
    # Compute memory segment size, including all NULLs.
    size = sum(len(arg) for arg in argl) + argc.value

    return argl, address, size


def read_byte(address):
    type_ = ctypes.POINTER(ctypes.c_char)
    return ctypes.cast(address, type_).contents.value[0]


def reverse_walk_memory(address, limit=8192):
    pos = address
    for i in range(limit):
        pos = address - i
        yield pos, read_byte(pos)


def reverse_find_nulstring(walker):
    # Search backward null-terminated strings.
    string = ''
    null = 0 if PY3 else '\x00'
    for addr, b in walker:
        if null == b:
            if string:
                yield addr + 1, string
            string = ''
        else:
            c = chr(b) if PY3 else b
            string = c + string


def find_stack_segment_from_maps(lines):
    for line in lines:
        if not line.endswith('[stack]\n'):
            continue
        # See proc(3) for a description of /proc/self/maps format.
        segment = line.split()
        address_range = segment[0]
        start, end = address_range.split('-')
        return int(start, base=16), int(end, base=16)
    else:
        raise Exception("Can't find stack segment")


def find_argv_memory_from_maps(maps, argv, environ=os.environ):
    # Find stack in /proc/self/maps, and walk backward until argv[0] is found.
    # Returns address of argv[0] and fixed argv when using python -c cmd form.
    stack_start, stack_end = find_stack_segment_from_maps(maps)
    argv = argv[:]
    remaining_argv = argv[:]
    # Stack ends with argv, environ and a single path. These are all null
    # terminated strings.
    found_environ = False
    walker = reverse_walk_memory(stack_end - 1, limit=stack_end - stack_start)
    for address, string in reverse_find_nulstring(walker):
        # Check if string is an entry from environ list.
        name, _, _ = string.partition('=')
        if name in environ:
            found_environ = True
            continue
        elif not found_environ:
            # We are after environ, just skip until we find an env var
            # declaration.
            continue
        elif string == remaining_argv[-1]:
            remaining_argv.pop()
            if not remaining_argv:
                # Ok we found all args!
                return argv, address
        elif remaining_argv[-1] == '__COMMAND_STRING__':
            logger.debug("Guessed command string: %r.", string)
            remaining_argv.pop()
            command_index = len(remaining_argv)
            argv[command_index] = string
        else:
            raise Exception("Can't find argv in stack segment")


class ProcTitleManager(object):
    def __init__(self, prefix):
        self.prefix = prefix
        self.address = self.size = None

    def setup(self):  # pragma: nocover
        try:
            argv, self.address, self.size = find_argv_memory_from_pythonapi()
            if sys.version_info > (2,):  # pragma: nocover_py2
                # On CPython3, PythonAPI returns a copy of argv. Find argv
                # address from /proc/self/maps.
                with open('/proc/self/maps') as fo:
                    argv, self.address = find_argv_memory_from_maps(
                        argv=argv, maps=fo)
                    self.size = sum(len(s) for s in argv) + len(argv)
            logger.debug("argv is at %#x, len=%d.", self.address, self.size)
        except Exception as e:
            self.address = self.size = False
            logger.debug("Failed to find argv memory segment: %s", e)

    def settitle(self, title):
        if not self.address:
            return

        # cf. https://dali.bo/tYIql
        title = self.prefix + title
        title = title.encode('utf-8')
        # Truncate title to fit in argv memory segment.
        title = title[:self.size - 1]
        # Pad title with NULL
        title = title.ljust(self.size, b'\0')
        # Overwrite argv segment with proc title
        dest = ctypes.create_string_buffer(title)
        ctypes.memmove(self.address, dest, self.size)
        return title

    def __call__(self, title):
        return self.settitle(title)


def test_main():  # pragma: nocover
    logging.basicConfig(level=logging.DEBUG)
    setproctitle = ProcTitleManager(prefix='temboard-process: ')
    setproctitle.setup()
    setproctitle('test process')
    with open('/proc/self/cmdline') as fo:
        cmdline = fo.read()
    logger.debug('/proc/self/cmdline is %r', cmdline)
    wanted = 'temboard-process: test process'
    assert wanted in cmdline

    if os.environ.get('WAIT'):
        logger.debug("Process title should be %r", wanted)
        print("Hit RET to terminate.", file=sys.stderr, end='')
        if PY3:
            input()
        else:
            raw_input()

    logger.info("OK")


if '__main__' == __name__:  # pragma: nocover
    test_main()



from datetime import datetime

from temboardagent.routing import RouteSet
from temboardagent.toolkit import taskmanager
from temboardagent.tools import validate_parameters
from temboardagent.types import T_OBJECTNAME

from . import functions


routes = RouteSet(prefix=b'/maintenance')
workers = taskmanager.WorkerSet()


@routes.get(b'', check_key=True)
def get_instance(http_context, app):
    with app.postgres.connect() as conn:
        instance = next(functions.get_instance(conn))

    with app.postgres.connect() as conn:
        rows = functions.get_databases(conn)

    databases = []
    for database in rows:
        database = dict(database)
        # we need to connect with a different database
        dbname = database['datname']
        with functions.get_postgres(app.config, dbname).connect() as conn:
            database.update(**functions.get_database(conn))
        databases.append(database)

    return {'instance': instance, 'databases': databases}


T_DATABASE_NAME = T_OBJECTNAME
T_SCHEMA_NAME = T_OBJECTNAME
T_TABLE_NAME = T_OBJECTNAME
T_INDEX_NAME = T_OBJECTNAME
T_VACUUM_MODE = b'(((^|,)(full|freeze|analyze))+$)'
T_TIMESTAMP_UTC = b'(^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$)'
T_OPERATION_ID = b'(^[0-9a-f]{8}$)'


@routes.get(b'/%s' % (T_DATABASE_NAME), check_key=True)
def get_database(http_context, app):
    dbname = http_context['urlvars'][0]
    with functions.get_postgres(app.config, dbname).connect() as conn:
        database = functions.get_database_size(conn)
        schemas = functions.get_schemas(conn)
    return dict(database, **{'schemas': schemas})


@routes.get(b'/%s/schema/%s' % (T_DATABASE_NAME, T_SCHEMA_NAME),
            check_key=True)
def get_schema(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    with functions.get_postgres(app.config, dbname).connect() \
            as conn:
        tables = functions.get_tables(conn, schema)
        indexes = functions.get_schema_indexes(conn, schema)
        schema = functions.get_schema(conn, schema)
    return dict(dict(tables, **indexes), **schema)


@routes.get(b'/%s/schema/%s/table/%s' % (T_DATABASE_NAME, T_SCHEMA_NAME,
                                         T_TABLE_NAME),
            check_key=True)
def get_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    table = http_context['urlvars'][2]

    with functions.get_postgres(app.config, dbname).connect() \
            as conn:
        ret = functions.get_table(conn, schema, table)
        ret.update(**functions.get_table_indexes(conn, schema, table))
        return ret


@routes.post(b'/%s/vacuum' % T_DATABASE_NAME)
def post_vacuum_database(http_context, app):
    dbname = http_context['urlvars'][0]
    return post_vacuum(app, http_context['post'], dbname)


@routes.post(b'/%s/schema/%s/table/%s/vacuum' % (T_DATABASE_NAME,
                                                 T_SCHEMA_NAME,
                                                 T_TABLE_NAME))
def post_vacuum_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    table = http_context['urlvars'][2]
    return post_vacuum(app, http_context['post'], dbname, schema, table)


def post_vacuum(app, post, dbname, schema=None, table=None):
    # Parameters format validation
    if 'datetime' in post:
        validate_parameters(post, [
            ('datetime', T_TIMESTAMP_UTC, False),
        ])
    dt = post.get('datetime', datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))
    if 'mode' in post:
        validate_parameters(post, [
            ('mode', T_VACUUM_MODE, False),
        ])
    mode = post.get('mode', '')

    with functions.get_postgres(app.config, dbname).connect() as conn:
        return functions.schedule_vacuum(conn, dbname, mode, dt, app,
                                         schema=schema, table=table)


@routes.get(b'/%s/vacuum/scheduled' % (T_DATABASE_NAME), check_key=True)
def scheduled_vacuum_database(http_context, app):
    dbname = http_context['urlvars'][0]
    return functions.list_scheduled_vacuum(app, dbname=dbname)


@routes.get(b'/%s/schema/%s/table/%s/vacuum/scheduled' % (T_DATABASE_NAME,
                                                          T_SCHEMA_NAME,
                                                          T_TABLE_NAME),
            check_key=True)
def scheduled_vacuum_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    table = http_context['urlvars'][2]
    return functions.list_scheduled_vacuum(app, dbname=dbname, schema=schema,
                                           table=table)


@routes.delete(b'/vacuum/' + T_OPERATION_ID)
def delete_vacuum(http_context, app):
    vacuum_id = http_context['urlvars'][0]
    return functions.cancel_scheduled_operation(vacuum_id, app)


@routes.get(b'/vacuum/scheduled', check_key=True)
def scheduled_vacuum(http_context, app):
    return functions.list_scheduled_vacuum(app)


@workers.register(pool_size=10)
def vacuum_worker(app, dbname, mode, schema=None, table=None):
    with functions.get_postgres(app.config, dbname).connect() \
            as conn:
        return functions.vacuum(conn, dbname, mode, schema, table)


@routes.post(b'/%s/analyze' % T_DATABASE_NAME)
def post_analyze_database(http_context, app):
    dbname = http_context['urlvars'][0]
    return post_analyze(app, http_context['post'], dbname)


@routes.post(b'/%s/schema/%s/table/%s/analyze' % (T_DATABASE_NAME,
                                                  T_SCHEMA_NAME,
                                                  T_TABLE_NAME))
def post_analyze_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    table = http_context['urlvars'][2]
    return post_analyze(app, http_context['post'], dbname, schema, table)


def post_analyze(app, post, dbname, schema=None, table=None):
    # Parameters format validation
    if 'datetime' in post:
        validate_parameters(post, [
            ('datetime', T_TIMESTAMP_UTC, False),
        ])
    dt = post.get('datetime', datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))

    with functions.get_postgres(app.config, dbname).connect() as conn:
        return functions.schedule_analyze(conn, dbname, dt, app,
                                          schema=schema, table=table)


@routes.get(b'/%s/analyze/scheduled' % (T_DATABASE_NAME), check_key=True)
def scheduled_analyze_database(http_context, app):
    dbname = http_context['urlvars'][0]
    return functions.list_scheduled_analyze(app, dbname=dbname)


@routes.get(b'/%s/schema/%s/table/%s/analyze/scheduled' % (T_DATABASE_NAME,
                                                           T_SCHEMA_NAME,
                                                           T_TABLE_NAME),
            check_key=True)
def scheduled_analyze_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    table = http_context['urlvars'][2]
    return functions.list_scheduled_analyze(app, dbname=dbname, schema=schema,
                                            table=table)


@routes.delete(b'/analyze/' + T_OPERATION_ID)
def delete_analyze(http_context, app):
    analyze_id = http_context['urlvars'][0]
    return functions.cancel_scheduled_operation(analyze_id, app)


@routes.get(b'/analyze/scheduled', check_key=True)
def scheduled_analyze(http_context, app):
    return functions.list_scheduled_analyze(app)


@workers.register(pool_size=10)
def analyze_worker(app, dbname, schema=None, table=None):
    with functions.get_postgres(app.config, dbname).connect() \
            as conn:
        return functions.analyze(conn, dbname, schema, table)


@routes.post(b'/%s/reindex' % (T_DATABASE_NAME))
def post_reindex_database(http_context, app):
    dbname = http_context['urlvars'][0]
    return post_reindex(app, http_context['post'], dbname)


@routes.post(b'/%s/schema/%s/table/%s/reindex' % (T_DATABASE_NAME,
                                                  T_SCHEMA_NAME,
                                                  T_TABLE_NAME))
def post_reindex_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    table = http_context['urlvars'][2]
    return post_reindex(app, http_context['post'], dbname, schema=schema,
                        table=table)


@routes.post(b'/%s/schema/%s/index/%s/reindex' % (T_DATABASE_NAME,
                                                  T_SCHEMA_NAME,
                                                  T_INDEX_NAME))
def post_reindex_index(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    index = http_context['urlvars'][2]
    return post_reindex(app, http_context['post'], dbname, schema=schema,
                        index=index)


def post_reindex(app, post, dbname, schema=None, table=None, index=None):
    # Parameters format validation
    if 'datetime' in post:
        validate_parameters(post, [
            ('datetime', T_TIMESTAMP_UTC, False),
        ])
    dt = post.get('datetime', datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))

    with functions.get_postgres(app.config, dbname).connect() as conn:
        return functions.schedule_reindex(
            conn, dbname, dt, app, schema=schema, table=table, index=index)


@routes.get(b'/%s/reindex/scheduled' % (T_DATABASE_NAME), check_key=True)
def scheduled_reindex_database(http_context, app):
    dbname = http_context['urlvars'][0]
    return functions.list_scheduled_reindex(app, dbname=dbname)


@routes.get(b'/%s/schema/%s/table/%s/reindex/scheduled' %
            (T_DATABASE_NAME, T_SCHEMA_NAME, T_TABLE_NAME),
            check_key=True)
def scheduled_reindex_table(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    return functions.list_scheduled_reindex(app, dbname=dbname, schema=schema)


@routes.get(b'/%s/schema/%s/reindex/scheduled' % (T_DATABASE_NAME,
                                                  T_SCHEMA_NAME),
            check_key=True)
def scheduled_reindex_index(http_context, app):
    dbname = http_context['urlvars'][0]
    schema = http_context['urlvars'][1]
    return functions.list_scheduled_reindex(app, dbname=dbname, schema=schema)


@routes.delete(b'/reindex/' + T_OPERATION_ID)
def delete_reindex(http_context, app):
    reindex_id = http_context['urlvars'][0]
    return functions.cancel_scheduled_operation(reindex_id, app)


@routes.get(b'/reindex/scheduled', check_key=True)
def scheduled_reindex(http_context, app):
    return functions.list_scheduled_reindex(app)


@workers.register(pool_size=10)
def reindex_worker(app, dbname, schema=None, table=None, index=None):
    with functions.get_postgres(app.config, dbname).connect() as conn:
        return functions.reindex(conn, dbname, schema, table, index)


class MaintenancePlugin:
    PG_MIN_VERSION = (90400, 9.4)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        self.app.router.add(routes)
        self.app.worker_pool.add(workers)

    def unload(self):
        self.app.worker_pool.remove(workers)
        self.app.router.remove(routes)



import sys

from ..usermgmt import hash_password


def display_usage():
    sys.stdout.write(
        "Usage: temboard-agent-password <username>:<password>\n"
        "\n"
        "Build an authentication string for temboard-agent based on "
        "username/password couple.\n"
        "\n"
        "Options:\n"
        "  -h, --help           show this help message and exit\n")
    exit(0)


def main():
    if len(sys.argv) != 2:
        display_usage()
    if sys.argv[1] == "-h" or sys.argv == "--help":
        display_usage()
    expl = str(sys.argv[1]).split(':')
    if len(expl) != 2:
        display_usage()
    sys.stdout.write("{}:{}\n".format(
        expl[0],
        hash_password(expl[0], expl[1]).decode('utf-8'))
    )
    exit(0)


if __name__ == '__main__':
    main()



import logging
import time

from temboardagent.routing import add_route
from temboardagent.errors import (
    HTTPError,
    NotificationError,
    SharedItem_exists,
    SharedItem_no_free_slot_left,
    SharedItem_not_found,
)
from temboardagent.sharedmemory import Session
from temboardagent.types import (
    T_PASSWORD,
    T_SESSIONID,
    T_USERNAME,
)
from temboardagent.tools import validate_parameters
from temboardagent.usermgmt import auth_user, gen_sessionid
from temboardagent.notification import NotificationMgmt, Notification
from temboardagent.inventory import SysInfo, PgInfo
from .version import __version__ as version


logger = logging.getLogger(__name__)


def check_sessionid(http_header, sessions):
    """
    Check X-Session is valid and the session exists.
    """
    validate_parameters(http_header,
                        [('X-Session', T_SESSIONID, False)])
    try:
        xsession = http_header['X-Session'].encode('utf-8')
        session = sessions.get_by_sessionid(xsession)
        session.time = time.time()
        username = session.username
        sessions.update(session)
        return username
    except SharedItem_not_found:
        raise HTTPError(401, "Invalid session.")


@add_route('POST', b'/login', check_session=False)
def login(http_context, app, sessions):
    post = http_context['post']
    # Add an unconditional sleeping time to reduce brute-force risks
    time.sleep(1)

    logger.info("Authenticating user: %s" % (post['username']))
    try:
        validate_parameters(post,
                            [('username', T_USERNAME, False),
                             ('password', T_PASSWORD, False)])
        auth_user(app.config.temboard['users'],
                  post['username'], post['password'])
    except HTTPError as e:
        logger.info("Authentication failed.")
        raise e
    try:
        session = sessions.get_by_username(post['username'])
        if not session:
            sessionid = gen_sessionid(post['username'])
            session = Session(sessionid.encode('utf-8'),
                              time.time(),
                              post['username'].encode('utf-8'))
            sessions.add(session)
        else:
            sessionid = session.sessionid
            session.time = time.time()
            sessions.update(session)
        try:
            NotificationMgmt.push(app.config,
                                  Notification(username=post['username'],
                                               message="Login"))
        except NotificationError as e:
            logger.exception(e)

        return {'session': sessionid}
    except (SharedItem_exists, SharedItem_no_free_slot_left) as e:
        logger.exception(e)
        raise HTTPError(500, "Internal error.")


@add_route('GET', b'/logout')
def logout(http_context, app, sessions):
    headers = http_context['headers']
    logger.info("Removing session: %s" % (headers['X-Session']))
    try:
        NotificationMgmt.push(app.config,
                              Notification(username=http_context['username'],
                                           message="Logout"))
    except NotificationError as e:
        logger.exception(e)

    try:
        sessions.delete(headers['X-Session'].encode('utf-8'))
        return {'logout': True}
    except (SharedItem_exists, SharedItem_no_free_slot_left) as e:
        logger.exception(e)
        raise HTTPError(500, "Internal error.")


@add_route('GET', b'/discover', check_session=False)
def get_discover(http_context, app, sessions):
    logger.info('Starting discovery.')

    discover = dict(
        hostname=None,
        cpu=None,
        memory_size=None,
        pg_port=app.config.postgresql['port'],
        pg_version=None,
        pg_version_summary=None,
        pg_data=None,
        plugins=[plugin for plugin in app.config.temboard['plugins']],
    )

    try:
        # Gather system informations
        sysinfo = SysInfo()
        hostname = sysinfo.hostname(app.config.temboard['hostname'])
        cpu = sysinfo.n_cpu()
        memory_size = sysinfo.memory_size()

    except (Exception, HTTPError) as e:
        logger.exception(str(e))
        logger.error('System discovery failed.')
        # We stop here if system information has not been collected
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")

    discover.update(
        hostname=hostname, cpu=cpu, memory_size=memory_size
    )

    try:
        with app.postgres.connect() as conn:
            pginfo = PgInfo(conn)
            discover.update(
                pg_block_size=int(pginfo.setting('block_size')),
                pg_version=pginfo.version()['full'],
                pg_version_summary=pginfo.version()['summary'],
                pg_data=pginfo.setting('data_directory')
            )

    except Exception as e:
        logger.exception(str(e))
        logger.error('Postgres discovery failed.')
        # Do not raise HTTPError, just keeping null values for Postgres
        # informations.

    logger.info('Discovery done.')
    logger.debug(discover)
    return discover


@add_route('GET', b'/profile')
def profile(http_context, app, sessions):
    headers = http_context['headers']
    logger.info("Get user profile.")
    try:
        xsession = headers['X-Session'].encode('utf-8')
        session = sessions.get_by_sessionid(xsession)
        logger.info("Done.")
        return {'username': session.username.decode('utf-8')}
    except SharedItem_not_found as e:
        logger.exception(e)
        logger.info("Failed.")
        raise HTTPError(401, "Invalid session.")


@add_route('GET', b'/notifications', check_key=True)
def notifications(http_context, app, sessions):
    logger.info("Get notifications.")
    try:
        notifications = NotificationMgmt.get_last_n(app.config, -1)
        logger.info("Done.")
        return list(notifications)
    except (NotificationError, Exception) as e:
        logger.exception(e)
        logger.info("Failed.")
        raise HTTPError(500, "Internal error.")


@add_route('GET', b'/status', check_session=False)
def get_status(http_context, app, sessions):
    logger.info('Starting /status.')
    try:
        start_datetime = time.strftime("%Y-%m-%dT%H:%M:%S%Z",
                                       app.start_datetime.timetuple())
        reload_datetime = time.strftime("%Y-%m-%dT%H:%M:%S%Z",
                                        app.start_datetime.timetuple())
        ret = dict(
            pid=app.pid,
            user=app.user,
            start_datetime=start_datetime,
            reload_datetime=reload_datetime,
            configfile=app.config.temboard.configfile,
            version=version,
        )
        logger.info('Ending /status.')
        return ret
    except (Exception, HTTPError) as e:
        logger.exception(str(e))
        logger.info('Status failed.')
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")



class UserError(Exception):
    def __init__(self, message, retcode=1):
        super(UserError, self).__init__(message)
        self.retcode = retcode


class StorageEngineError(Exception):
    pass



import logging
import re
from textwrap import dedent

import psycopg2.extensions
from psycopg2 import connect
from psycopg2.extensions import connection
from psycopg2.extras import RealDictCursor

from .errors import UserError


logger = logging.getLogger(__name__)


# See https://www.psycopg.org/docs/faq.html#faq-float
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    'DEC2FLOAT',
    lambda value, curs: float(value) if value is not None else None)
psycopg2.extensions.register_type(DEC2FLOAT)


def pg_escape(in_string, escapee_char=r"'"):
    out_string = ''
    out_string += escapee_char
    out_string += re.sub(escapee_char, escapee_char * 2, in_string)
    out_string += escapee_char
    return out_string


class ConnectionHelper(connection):
    def execute(self, query, vars=None):
        with self.cursor() as cur:
            cur.execute(dedent(query), vars)

    def query(self, query, vars=None):
        with self.cursor() as cur:
            cur.execute(dedent(query), vars)
            yield from cur

    def queryone(self, query, vars=None):
        with self.cursor() as cur:
            cur.execute(dedent(query), vars)
            return cur.fetchone()

    def query_scalar(self, query, vars=None):
        row = self.queryone(query, vars)
        if row is None:
            return
        else:
            return next(iter(row.values()))


class ConnectionManager:
    def __init__(self, postgres, app):
        self.postgres = postgres
        self.app = app
        self.conn = None

    def open(self):
        logger.debug(
            "Opening Postgres connexion to database %s.",
            self.postgres.dbname)
        try:
            self.conn = connect(
                host=self.postgres.host,
                port=self.postgres.port,
                user=self.postgres.user,
                password=self.postgres.password,
                database=self.postgres.dbname,
                connection_factory=ConnectionHelper,
                cursor_factory=RealDictCursor,
                application_name='temboard-agent',
            )
            self.conn.set_session(autocommit=True)
            if self.app is not None:
                self.app.check_compatibility(self.conn.server_version)
        except Exception as e:
            raise UserError("Failed to connect to Postgres: %s" % e)
        return self.conn

    def close(self):
        logger.debug("Closing connection to %s.", self.postgres.dbname)
        self.conn.close()
        self.conn = None

    def __enter__(self):
        if self.conn is None:
            self.open()
        return self.conn

    def __exit__(self, *a):
        self.close()


class Postgres:
    def __init__(
            self, host=None, port=5432, user=None, password=None, dbname=None,
            app=None,
            **kw):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        # Compat with conninfo dict.
        if 'database' in kw:
            dbname = kw['database']
        self.dbname = dbname
        self.app = app
        self._server_version = None

    def __repr__(self):
        return '<{} on {}@{}:{}/{}>'.format(
            self.__class__.__name__,
            self.user, self.host, self.port, self.dbname,
        )

    def connect(self):
        return ConnectionManager(self, self.app)

    def fetch_version(self):
        if self._server_version is None:
            with self.connect() as conn:
                self._server_version = conn.server_version
        return self._server_version

    def copy(self, **kw):
        defaults = dict(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            dbname=self.dbname,
            app=self.app,
        )
        kw = dict(defaults, **kw)
        return self.__class__(**kw)


class ConnectionPool:
    def __init__(self, **kw):
        if 'database' in kw:
            kw['dbname'] = kw.pop('database')
        self.conn_kw = kw
        self.pool = dict()

    def get(self, dbname=None):
        conn_kw = self.conn_kw.copy()
        if dbname:
            conn_kw['dbname'] = dbname
        dbname = conn_kw['dbname']
        try:
            return self.pool[dbname]
        except KeyError:
            conn = Postgres(**conn_kw).connect()
            return self.pool.setdefault(dbname, conn.open())

    def __enter__(self):
        return self

    def __exit__(self, *_):
        for conn in self.pool.values():
            conn.close()
        self.pool = dict()



import logging
import time

import json
import sys
from http.server import BaseHTTPRequestHandler, HTTPServer
from socket import error as SocketError
from socketserver import ThreadingMixIn
from urllib.parse import urlparse, parse_qs, unquote_plus
import ssl

from temboardagent.routing import get_routes
from temboardagent.errors import HTTPError
from temboardagent.tools import JSONEncoder
from temboardagent import __version__ as temboard_version
from .sharedmemory import Sessions
from .api import check_sessionid
from .errors import UserError
from .toolkit.services import Service


logger = logging.getLogger(__name__)


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """ Handle requests in a separate thread. """


class RequestHandler(BaseHTTPRequestHandler):
    """
    HTTP request handler.
    """
    def __init__(self, app, sessions, *args, **kwargs):
        """
        Constructor.
        """
        # Sessions array in shared memory.
        self.sessions = sessions
        # Application instance.
        self.app = app
        # HTTP server version.
        self.server_version = "temboard-agent/%s" % temboard_version
        # HTTP request method
        self.http_method = None
        # HTTP query.
        self.query = None
        # HTTP POST content in json format.
        self.post_json = None
        # Call HTTP request handler constructor.
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def do_GET(self):
        """
        Handle HTTP GET requests.
        """
        self.http_method = 'GET'
        self.response()

    def do_POST(self,):
        """
        Handle HTTP POST requests.
        """
        self.http_method = 'POST'
        self.response()

    def do_PUT(self,):
        """
        Handle HTTP PUT requests.
        """
        self.http_method = 'PUT'
        # Nothing to do for now.
        pass

    def do_DELETE(self,):
        """
        Handle HTTP DELETE requests.
        """
        self.http_method = 'DELETE'
        self.response()

    def do_OPTIONS(self,):
        self.send_response(200, "OK")
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods',
                         'POST, GET, DELETE, OPTIONS')
        self.send_header('Access-Control-Allow-Headers',
                         "X-Requested-With, X-Session, Content-Type")
        self.send_header('Access-Control-Max-Age', '1728000')
        self.end_headers()
        logger.info(self.headers.dict['origin'])

    def log_message(self, format, *args):
        """
        Overrides log_message() for HTTP requests logging with our own logger.
        """
        logger.info("client: %s request: %s"
                    % (self.address_string(), format % args))

    def handle_one_request(self, *a, **kw):
        self.start_time = time.time()
        return BaseHTTPRequestHandler.handle_one_request(self, *a, **kw)

    def request_time(self):
        return 1000. * (time.time() - self.start_time)

    def log_request(self, code='-', size='-'):
        """Log an accepted request.

        This is called by send_response().

        """
        if hasattr(code, 'value'):
            code = code.value
        self.log_message(
            '"%s" %s %s %.2fms',
            self.requestline, str(code), str(size), self.request_time(),
        )

    def response(self):
        """
        In charge to call the main routing function and to return its results
        as a valid HTTP response.
        """
        try:
            (code, message) = self.route_request()
        except HTTPError as e:
            logger.exception(e.message)
            logger.error(e.message)
            code = e.code
            message = e.message
        except UserError as e:
            msg = str(e)
            logger.exception(msg)
            logger.error(msg)
            code = 500
            message = {'error': msg}
        except Exception as e:
            logger.exception(str(e))
            logger.error("Internal error")
            # This is an unknown error. Just inform there is an internal error.
            code = 500
            message = {'error': "Internal error."}

        try:
            # Try to send the response
            self.send_response(int(code))
            self.send_header('Access-Control-Allow-Origin', '*')
            self.send_header('Content-type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps(message, cls=JSONEncoder)
                             .encode('utf-8'))
        except Exception as e:
            logger.exception(str(e))
            logger.error("Could not send response")

    def get_route(self, method, path):
        # Returns the right route according to method/path
        s_path = path.split(b'/')[1:]
        root = s_path[0]
        for route in get_routes():
            # Check that HTTP method and url root are matching.
            if not (route['http_method'] == self.http_method and
                    route['root'] == root):
                continue

            p = 0
            # Check each element in the path.
            for elt in s_path:
                try:
                    if type(route['splitpath'][p]) not in (str, bytes):
                        # Then this is a regular expression.
                        res = route['splitpath'][p].match(elt.decode('utf-8'))
                        if not res:
                            break
                    else:
                        if route['splitpath'][p] != elt:
                            break
                except IndexError:
                    break
                p += 1
            if p == len(s_path) == len(route['splitpath']):
                logger.debug(route)
                return route
        raise HTTPError(404, 'URL not found.')

    def parse_path(self, path, route):
        # Parse an URL path when route's path contains regepx
        p = 0
        urlvars = list()
        for elt in path.split(b'/')[1:]:
            if type(route['splitpath'][p]) not in (str, bytes):
                # Then this is a regular expression.
                res = route['splitpath'][p].match(elt.decode('utf-8'))
                if res is not None:
                    # If the regexp matches, we want to get the
                    # value and append it in urlvars.
                    urlvars.append(unquote_plus(res.group(1)))
            p += 1
        return urlvars

    def route_request(self,):
        """
        Main function in charge to route the incoming HTTP request to the right
        function.
        """
        # Let's parse and prepare url, path, query etc..
        up = urlparse(self.path, 'http')
        path = up.path
        splitpath = path.split('/')
        if len(splitpath) == 1:
            raise HTTPError(404, 'Not found.')
        self.query = parse_qs(up.query)

        # Get the route
        route = self.get_route(self.http_method, path.encode('utf-8'))
        # Parse URL path
        urlvars = self.parse_path(path.encode('utf-8'), route)
        post_raw = None
        # Load POST content if any
        if self.http_method == 'POST':
            # TODO: raise an HTTP error if the content-length is
            # too large.
            try:
                post_raw = self.rfile.read(int(self.headers['Content-Length']))
            except Exception as e:
                logger.exception(str(e))
                logger.debug(self.headers)
                logger.error('Unable to read post data')
                raise HTTPError(400, 'Unable to read post data')

        username = None
        checked = False

        # Authentication checking out

        # 1. Try the auth' by key: if this method is available for this API
        # and 'key' arg exists.
        key = self.headers.get('X-TemBoard-Agent-Key')
        if key:
            logger.debug("Authenication by key from header.")
        elif 'key' in self.query:
            # TODO: Remove auth from query in 8.0
            key = self.query['key'][0]
            logger.debug("Authentication by key from argument.")

        if route['check_key'] and key:
            if self.app.config.temboard.key is None:
                raise HTTPError(401, "Authentication key not configured")
            if key != self.app.config.temboard.key:
                raise HTTPError(401, "Invalid key")
            checked = True

        # 2. Check session ID if available and not previously auth'd by key
        if not checked and route['check_session']:
            username = check_sessionid(self.headers, self.sessions)
            checked = True

        # 3. At this point, if not yet checked out and auth' by key is
        # available then we need to raise an error because 'key' arg hasn't
        # been passed and auth' by key is the only available method.
        if not checked and route['check_key']:
            raise HTTPError(401, "Missing key")

        try:
            # Load POST content expecting it is in JSON format.
            if self.http_method == 'POST':
                self.post_json = json.loads(post_raw.decode('utf-8'))
        except Exception as e:
            logger.exception(str(e))
            logger.error('Invalid json format')
            raise HTTPError(400, 'Invalid json format')

        http_context = dict(
            headers=self.headers,
            query=self.query,
            post=self.post_json,
            urlvars=urlvars,
            username=username,
        )

        # Handle the request
        func = getattr(sys.modules[route['module']], route['function'])
        if route['module'] == 'temboardagent.api':
            # some core APIs need to deal with sessions
            return (200, func(http_context, self.app, self.sessions))
        else:
            # plugin
            return (200, func(http_context, self.app))


class HTTPDService(Service):
    def setup(self):
        self.sessions = Sessions(size=100)
        try:
            self.httpd = ThreadedHTTPServer(
                (self.app.config.temboard.address,
                 self.app.config.temboard.port),
                self.handle_request)
        except SocketError as e:
            raise UserError("Failed to start HTTPS server: {}.".format(e))
        try:
            logger.debug(
                "Using SSL key %s.", self.app.config.temboard.ssl_key_file)
            logger.debug(
                "Using SSL certificate %s.",
                self.app.config.temboard.ssl_cert_file)
            self.httpd.socket = ssl.wrap_socket(
                self.httpd.socket,
                keyfile=self.app.config.temboard.ssl_key_file,
                certfile=self.app.config.temboard.ssl_cert_file,
                server_side=True,
            )
        except Exception as e:
            raise UserError("Failed to setup SSL: {}.".format(e))
        self.httpd.timeout = 1

    def serve1(self):
        self.httpd.handle_request()
        self.sessions.purge_expired(3600, logger, self.app.config)

    def handle_request(self, *args):
        return RequestHandler(self.app, self.sessions, *args)



from multiprocessing import Lock
from multiprocessing.sharedctypes import Array
from ctypes import (Structure, c_char, c_double)
import time

from .errors import (
    SharedItem_not_found, SharedItem_exists,
    SharedItem_bad_type_size,
    SharedItem_no_free_slot_left,
    NotificationError,
)
from .types import T_SESSIONID_SIZE, T_USERNAME_SIZE
from .notification import NotificationMgmt, Notification


class Session(Structure):
    """
    Session structure. The Session Array is shared by the HTTP server process
    and the scheduler process.
    """
    _fields_ = [
        # Session ID.
        ('sessionid', c_char * T_SESSIONID_SIZE),
        # Last update timestamp.
        ('time', c_double),
        # Username.
        ('username', c_char * T_USERNAME_SIZE)
    ]


class Sessions:
    """
    Sessions object.
    """
    def __init__(self, size=100):
        # Lock handler.
        self.lock = Lock()
        # Array size.
        self.size = size
        # Array of Session.
        self.sessions = Array(Session, self.size, lock=self.lock)

    def get_by_sessionid(self, sessionid):
        """
        Returns a Session by a sessionid.
        """
        for session in self.sessions:
            if session.sessionid == sessionid:
                return session
        raise SharedItem_not_found()

    def get_by_username(self, username):
        """
        Returns a Session by a username.
        """
        for session in self.sessions:
            if session.username == username:
                return session

    def add(self, session):
        """
        Add a new Session.
        """
        for i in range(0, self.size):
            if self.sessions[i].sessionid == session.sessionid:
                raise SharedItem_exists()
            if self.sessions[i].username == session.username:
                self.sessions[i] = Session()
        for i in range(0, self.size):
            if self.sessions[i].sessionid == b'':
                self.sessions[i] = session
                return
        raise SharedItem_no_free_slot_left()

    def update(self, session):
        """
        Modify a Session.
        """
        for i in range(0, self.size):
            if self.sessions[i].sessionid == session.sessionid:
                self.sessions[i] = session
                return
        raise SharedItem_not_found()

    def delete(self, sessionid):
        """
        Remove a Session.
        """
        for i in range(0, self.size):
            if self.sessions[i].sessionid == sessionid:
                self.sessions[i] = Session()
                return
        raise SharedItem_not_found()

    def add_tuple(self, t_session):
        """
        Convert to Session the given tuple and add it.
        """
        session = Session()
        if len(t_session) != len(session._fields_):
            raise SharedItem_bad_type_size()
        session.sessionid = t_session[0][:T_SESSIONID_SIZE]
        session.time = t_session[1]
        session.username = t_session[2][:T_USERNAME_SIZE]
        self.add(session)

    def purge_expired(self, ttl, logger, config):
        """
        Remove old Session when the Session last update time + TTL is prior to
         current timestamp.
        """
        for i in range(0, self.size):
            if len(self.sessions[i].sessionid) == T_SESSIONID_SIZE and \
               (self.sessions[i].time + ttl) < time.time():
                logger.info("Session with sessionid=%s expired." %
                            (self.sessions[i].sessionid))
                try:
                    NotificationMgmt.push(config, Notification(
                        username=self.sessions[i].username,
                        message="Session expired"))
                except NotificationError as e:
                    logger.error(e.message)

                self.sessions[i] = Session()



from datetime import datetime
import time
import logging
import json

from ...toolkit import taskmanager
from ...routing import RouteSet
from ...toolkit.configuration import OptionSpec
from ...toolkit.validators import commalist
from ...tools import now, validate_parameters
from ...inventory import SysInfo
from ... import __version__ as __VERSION__
from ...errors import HTTPError as TemboardHTTPError
from ...postgres import ConnectionPool

from . import db
from .inventory import host_info, instance_info
from .probes import (
    load_probes,
    probe_bgwriter,
    probe_blocks,
    probe_cpu,
    probe_db_size,
    probe_filesystems_size,
    probe_heap_bloat,
    probe_btree_bloat,
    probe_loadavg,
    probe_locks,
    probe_memory,
    probe_process,
    probe_replication_lag,
    probe_replication_connection,
    probe_sessions,
    probe_temp_files_size_delta,
    probe_tblspc_size,
    probe_wal_files,
    probe_xacts,
    run_probes,
)
from .output import remove_passwords

logger = logging.getLogger(__name__)
workers = taskmanager.WorkerSet()
routes = RouteSet(prefix=b'/monitoring')

T_TIMESTAMP_UTC = b'(^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$)'
T_LIMIT = b'(^[0-9]+$)'


@routes.get(b'/probe/sessions', check_key=True)
def get_probe_sessions(http_context, app):
    return api_run_probe(probe_sessions(app.config.monitoring), app.config)


@routes.get(b'/probe/xacts', check_key=True)
def get_probe_xacts(http_context, app):
    return api_run_probe(probe_xacts(app.config.monitoring), app.config)


@routes.get(b'/probe/locks', check_key=True)
def get_probe_locks(http_context, app):
    return api_run_probe(probe_locks(app.config.monitoring), app.config)


@routes.get(b'/probe/blocks', check_key=True)
def get_probe_blocks(http_context, app):
    return api_run_probe(probe_blocks(app.config.monitoring), app.config)


@routes.get(b'/probe/bgwriter', check_key=True)
def get_probe_bgwriter(http_context, app):
    return api_run_probe(probe_bgwriter(app.config.monitoring), app.config)


@routes.get(b'/probe/db_size', check_key=True)
def get_probe_db_size(http_context, app):
    return api_run_probe(probe_db_size(app.config.monitoring), app.config)


@routes.get(b'/probe/tblspc_size', check_key=True)
def get_probe_tblspc_size(http_context, app):
    return api_run_probe(probe_tblspc_size(app.config.monitoring), app.config)


@routes.get(b'/probe/filesystems_size', check_key=True)
def get_probe_filesystems_size(http_context, app):
    return api_run_probe(probe_filesystems_size(app.config.monitoring),
                         app.config)


@routes.get(b'/probe/cpu', check_key=True)
def get_probe_cpu(http_context, app):
    return api_run_probe(probe_cpu(app.config.monitoring), app.config)


@routes.get(b'/probe/process', check_key=True)
def get_probe_process(http_context, app):
    return api_run_probe(probe_process(app.config.monitoring), app.config)


@routes.get(b'/probe/memory', check_key=True)
def get_probe_memory(http_context, app):
    return api_run_probe(probe_memory(app.config.monitoring), app.config)


@routes.get(b'/probe/loadavg', check_key=True)
def get_probe_loadavg(http_context, app):
    return api_run_probe(probe_loadavg(app.config.monitoring), app.config)


@routes.get(b'/probe/wal_files', check_key=True)
def get_probe_wal_files(http_context, app):
    return api_run_probe(probe_wal_files(app.config.monitoring), app.config)


@routes.get(b'/probe/replication_lag', check_key=True)
def get_probe_replication_lag(http_context, app):
    return api_run_probe(probe_replication_lag(app.config.monitoring),
                         app.config)


@routes.get(b'/probe/temp_files_size_delta', check_key=True)
def get_probe_temp_files_size_delta(http_context, app):
    return api_run_probe(probe_temp_files_size_delta(app.config.monitoring),
                         app.config)


@routes.get(b'/probe/replication_connection', check_key=True)
def get_probe_replication_connection(http_context, app):
    return api_run_probe(probe_replication_connection(app.config.monitoring),
                         app.config)


@routes.get(b'/probe/heap_bloat', check_key=True)
def get_probe_heap_bloat(http_context, app):
    return api_run_probe(probe_heap_bloat(app.config.monitoring), app.config)


@routes.get(b'/probe/btree_bloat', check_key=True)
def get_probe_btree_bloat(http_context, app):
    return api_run_probe(probe_btree_bloat(app.config.monitoring), app.config)


def api_run_probe(probe_instance, config):
    """
    Run a probe instance.
    """
    # Validate connection information from the config, and ensure
    # the instance is available
    conninfo = dict(
        host=config.postgresql.host,
        port=config.postgresql.port,
        user=config.postgresql.user,
        database=config.postgresql.dbname,
        password=config.postgresql.password,
        dbnames=config.monitoring.dbnames,
        instance=config.postgresql.instance,
    )
    sysinfo = SysInfo()
    hostname = sysinfo.hostname(config.temboard.hostname)
    with ConnectionPool(**conninfo) as pool:
        instance = instance_info(pool, conninfo, hostname)
        # Set home path
        probe_instance.set_home(config.temboard.home)
        # Gather the data from probes
        return run_probes([probe_instance], pool, [instance], delta=False)


@routes.get(b'/history', check_key=True)
def get_monitoring(http_context, app):
    """Monitoring root API aims to query metrics history.
    Data are sorted by collect timestamp, in ascending order. By default, only
    the most fresh record set is returned. The query parameter 'start' can be
    used as lower bound and must be expressed as a UTC timestamp formatted
    using ISO8601 norm with a terminal 'Z' character. To limit the number of
    returned records to N, the query parameter 'limit' can be used and set to
    N. 'limit' default value is 50, meaning that the maximum number of record
    set this API returns by default is 50.
    """

    # Default values
    start_timestamp = None
    limit = 50

    if 'start' in http_context['query']:
        # Validate start parameter
        validate_parameters(http_context['query'], [
            ('start', T_TIMESTAMP_UTC, True),
        ])
        # Convert it to epoch
        try:
            start_timestamp = (
                datetime.strptime(
                    http_context['query']['start'][0], "%Y-%m-%dT%H:%M:%SZ"
                ) - datetime(1970, 1, 1)
            ).total_seconds()
        except ValueError:
            raise TemboardHTTPError(406, "Invalid timestamp")

    if 'limit' in http_context['query']:
        # Validate limit parameter
        validate_parameters(http_context['query'], [
            ('limit', T_LIMIT, True),
        ])
        limit = int(http_context['query']['limit'][0])

    return [
        json.loads(metric[1]) for metric in db.get_metrics(
            app.config.temboard.home,
            'monitoring.db',
            start_timestamp=start_timestamp,
            limit=limit
        )
    ]


@routes.get(b'/config', check_key=True)
def get_config(http_context, app):
    """Returns monitoring plugin configuration.
    """
    return dict(
        dbnames=app.config.monitoring.dbnames,
        scheduler_interval=app.config.monitoring.scheduler_interval,
    )


@workers.register(pool_size=1)
def monitoring_collector_worker(app):
    """
    Run probes and push collected metrics in a queue.
    """
    logger.info("Starting monitoring collector.")
    config = app.config
    conninfo = dict(
        host=config.postgresql.host,
        port=config.postgresql.port,
        user=config.postgresql.user,
        database=config.postgresql.dbname,
        password=config.postgresql.password,
        dbnames=config.monitoring.dbnames,
        instance=config.postgresql.instance,
    )

    logger.info("Gathering host information.")
    system_info = host_info(config.temboard.hostname)
    logger.info("Load the probes to run.")
    probes = load_probes(
        config.monitoring,
        config.temboard.home
    )

    with ConnectionPool(**conninfo) as pool:
        instance = instance_info(pool, conninfo, system_info['hostname'])
        data = run_probes(probes, pool, [instance])

    # Prepare and send output
    output = dict(
        datetime=now(),
        hostinfo=system_info,
        instances=remove_passwords([instance]),
        data=data,
        version=__VERSION__,
    )
    logger.info("Add data to metrics table.")
    db.add_metric(
        config.temboard.home,
        'monitoring.db',
        time.time(),
        output
    )

    logger.info("Collect done.")

    try:
        logger.debug("temboard_agent_version=%s", __VERSION__)
        logger.debug("hostinfo=%s", system_info)
        for record in iter_metrics_for_logfmt(data):
            # up=1 is a marker to grep logfmt lines
            logger.debug(
                "up=1 %s",
                " ".join('%s=%s' % i for i in record.items()),
            )
    except Exception:
        logger.exception("Failed to log metrics.")


def iter_metrics_for_logfmt(data):
    # Generates a flat sequence of record dict containing key value for logfmt
    # printing. See perfui/ in temboard project to analyze such data.
    for k, v in data.items():
        for vv in v:
            record = dict()
            for kkk, vvv in vv.items():
                if hasattr(vvv, 'isoformat'):
                    vvv = vvv.isoformat(sep='T')
                if kkk in ('datetime', 'port', 'cpu', 'measure_interval'):
                    continue
                if kkk in ('dbname', 'spcname') or k in ('loadavg', 'memory'):
                    record[kkk] = vvv
                elif k.endswith('_size') and kkk == 'size':
                    record[k] = vvv
                elif k == 'lag':
                    record[k] = vvv
                else:
                    record['%s_%s' % (k, kkk)] = vvv
            yield record


class MonitoringPlugin:
    PG_MIN_VERSION = (90400, 9.4)
    s = 'monitoring'
    option_specs = [
        OptionSpec(s, 'dbnames', default='*', validator=commalist),
        OptionSpec(s, 'scheduler_interval', default=60, validator=int),
        OptionSpec(s, 'probes', default='*', validator=commalist),
    ]
    del s

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.option_specs)

    def bootstrap(self):
        db.bootstrap(self.app.config.temboard.home, 'monitoring.db')

    def load(self):
        self.app.router.add(routes)
        self.app.worker_pool.add(workers)
        logger.debug(
            "Schedule metric collect every %s seconds.",
            self.app.config.monitoring.scheduler_interval,
        )
        workers.schedule(
            id='monitoring_collector',
            redo_interval=self.app.config.monitoring.scheduler_interval,
        )(monitoring_collector_worker)
        self.app.scheduler.add(workers)

    def unload(self):
        self.app.scheduler.remove(workers)
        self.app.worker_pool.remove(workers)
        self.app.router.remove(routes)
        self.app.config.remove_specs(self.option_specs)



from .process import Process, memory_total_size
import time
from resource import getpagesize
from temboardagent.notification import NotificationMgmt, Notification
from temboardagent.tools import validate_parameters
from temboardagent.types import T_PID
from temboardagent.errors import NotificationError


def get_activity(conn):
    """
    Returns PostgreSQL backend list based on pg_stat_activity view.
    For each backend (process) we need to compute: CPU and mem. usage, I/O
    infos.
    """
    mem_total = memory_total_size()
    page_size = getpagesize()
    if conn.server_version >= 90600 and conn.server_version < 100000:
        query = """
SELECT
  pg_stat_activity.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.client_addr AS client,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric, 2)::FLOAT AS duration,
  CASE WHEN pg_stat_activity.wait_event_type IS
    DISTINCT FROM 'Lock' THEN 'N' ELSE 'Y' END AS wait,
  pg_stat_activity.usename AS user,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query
FROM
  pg_stat_activity
WHERE
  pid <> pg_backend_pid()
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
        """
    elif conn.server_version < 90600:
        query = """
SELECT
  pg_stat_activity.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.client_addr AS client,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric, 2)::FLOAT AS duration,
  CASE WHEN pg_stat_activity.waiting = 't' THEN 'Y' ELSE 'N' END AS wait,
  pg_stat_activity.usename AS user,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query
FROM
  pg_stat_activity
WHERE
  pid <> pg_backend_pid()
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
        """
    elif conn.server_version >= 100000:
        query = """
SELECT
  pg_stat_activity.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.client_addr AS client,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric, 2)::FLOAT AS duration,
  CASE WHEN pg_stat_activity.wait_event_type IS
    DISTINCT FROM 'Lock' THEN 'N' ELSE 'Y' END AS wait,
  pg_stat_activity.usename AS user,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query
FROM
  pg_stat_activity
WHERE
  pid <> pg_backend_pid()
  AND backend_type = 'client backend'
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
        """

    backend_list = []
    for row in conn.query(query):
        try:
            backend_list.append({
                'pid': row['pid'],
                'database': row['database'],
                'client': row['client'],
                'duration': row['duration'],
                'wait': row['wait'],
                'user': row['user'],
                'state': row['state'],
                'query': row['query'],
                'process': Process(row['pid'], mem_total, page_size)})
        except Exception:
            pass

    time.sleep(0.1)
    final_backend_list = []
    for row in backend_list:
        try:
            (read_s, write_s) = row['process'].io_usage()
            if row['duration'] < 0:
                row['duration'] = 0
            final_backend_list.append({
                'pid': row['pid'],
                'database': row['database'],
                'client': row['client'],
                'duration': row['duration'],
                'wait': row['wait'],
                'user': row['user'],
                'state': row['state'],
                'query': row['query'],
                'iow': row['process'].iow,
                'read_s': read_s,
                'write_s': write_s,
                'cpu': row['process'].cpu_usage(),
                'memory': row['process'].mem_usage()
            })
        except Exception:
            pass
    return {'rows': final_backend_list}


def post_activity_kill(conn, config, http_context):
    """
    Kill (using pg_terminate_backend()) processes based on a given backend PID
    list.
    """
    validate_parameters(http_context['post'], [
        ('pids', T_PID, True)
    ])
    ret = {'backends': []}
    for pid in http_context['post']['pids']:
        killed = conn.query_scalar(
            "SELECT pg_terminate_backend(%s) AS killed" % (pid))
        # Push a notification.
        try:
            NotificationMgmt.push(
                config,
                Notification(
                    username=http_context['username'],
                    message="Backend %s terminated" % (pid)
                )
            )
        except (NotificationError, Exception):
            pass

        ret['backends'].append({
            'pid': pid,
            'killed': killed,
        })
    return ret


def get_activity_waiting(conn):
    """
    Returns the list of waiting (on lock) queries.
    """
    mem_total = memory_total_size()
    page_size = getpagesize()

    query = """
SELECT
  pg_locks.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.usename AS user,
  pg_locks.mode AS mode,
  pg_locks.locktype AS type,
  COALESCE(pg_locks.relation::regclass::text, ' ') AS relation,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric,2)::FLOAT AS duration,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query
FROM
  pg_catalog.pg_locks JOIN pg_catalog.pg_stat_activity
    ON (pg_catalog.pg_locks.pid = pg_catalog.pg_stat_activity.pid)
WHERE
  NOT pg_catalog.pg_locks.granted
  AND pg_catalog.pg_stat_activity.pid <> pg_backend_pid()
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
    """
    backend_list = []
    for row in conn.query(query):
        try:
            backend_list.append({
                'pid': row['pid'],
                'database': row['database'],
                'user': row['user'],
                'mode': row['mode'],
                'type': row['type'],
                'relation': row['relation'],
                'duration': row['duration'],
                'state': row['state'],
                'query': row['query'],
                'process': Process(row['pid'], mem_total, page_size)})
        except Exception:
            pass

    time.sleep(0.1)
    final_backend_list = []
    for row in backend_list:
        try:
            (read_s, write_s) = row['process'].io_usage()
            if row['duration'] < 0:
                row['duration'] = 0
            final_backend_list.append({
                'pid': row['pid'],
                'database': row['database'],
                'user': row['user'],
                'mode': row['mode'],
                'type': row['type'],
                'relation': row['relation'],
                'duration': row['duration'],
                'state': row['state'],
                'query': row['query'],
                'iow': row['process'].iow,
                'read_s': read_s,
                'write_s': write_s,
                'cpu': row['process'].cpu_usage(),
                'memory': row['process'].mem_usage()
            })
        except Exception:
            pass
    return {'rows': final_backend_list}


def get_activity_blocking(conn):
    """
    Returns the list of blocking (lock) queries.
    """
    mem_total = memory_total_size()
    page_size = getpagesize()

    query = """
SELECT
  pid,
  datname AS database,
  usename AS user,
  COALESCE(relation::text, ' ') AS relation,
  mode,
  locktype AS type,
  duration,
  state,
  query
FROM (
  SELECT
    blocking.pid,
    pg_stat_activity.query,
    blocking.mode,
    pg_stat_activity.datname,
    pg_stat_activity.usename,
    blocking.locktype,
    round(EXTRACT(epoch FROM (NOW()
      - pg_stat_activity.query_start))::numeric,2)::FLOAT AS duration,
    blocking.relation::regclass AS relation,
    pg_stat_activity.state AS state
  FROM
    pg_locks AS blocking
    JOIN (SELECT transactionid FROM pg_locks WHERE NOT granted) AS blocked
      ON (blocking.transactionid = blocked.transactionid)
    JOIN pg_stat_activity
      ON (blocking.pid = pg_stat_activity.pid)
  WHERE
    blocking.granted
  UNION ALL
  SELECT
    blocking.pid,
    pg_stat_activity.query,
    blocking.mode,
    pg_stat_activity.datname,
    pg_stat_activity.usename,
    blocking.locktype,
    round(EXTRACT(epoch FROM (NOW()
      - pg_stat_activity.query_start))::numeric,2)::FLOAT AS duration,
    blocking.relation::regclass AS relation,
    pg_stat_activity.state AS state
  FROM
    pg_locks AS blocking
    JOIN (SELECT database, relation, mode FROM pg_locks WHERE NOT granted
        AND relation IS NOT NULL) AS blocked
      ON (blocking.database = blocked.database
        AND blocking.relation = blocked.relation)
    JOIN pg_stat_activity
      ON (blocking.pid = pg_stat_activity.pid)
  WHERE
    blocking.granted
) AS sq
GROUP BY pid, query, mode, locktype, duration, datname, usename, relation,
  state
ORDER BY duration DESC
    """
    backend_list = []
    for row in conn.query(query):
        try:
            backend_list.append({
                'pid': row['pid'],
                'database': row['database'],
                'user': row['user'],
                'mode': row['mode'],
                'type': row['type'],
                'relation': row['relation'],
                'duration': row['duration'],
                'state': row['state'],
                'query': row['query'],
                'process': Process(row['pid'], mem_total, page_size)})
        except Exception:
            pass

    time.sleep(0.1)
    final_backend_list = []
    for row in backend_list:
        try:
            (read_s, write_s) = row['process'].io_usage()
            if row['duration'] < 0:
                row['duration'] = 0
            final_backend_list.append({
                'pid': row['pid'],
                'database': row['database'],
                'user': row['user'],
                'mode': row['mode'],
                'type': row['type'],
                'relation': row['relation'],
                'duration': row['duration'],
                'state': row['state'],
                'query': row['query'],
                'iow': row['process'].iow,
                'read_s': read_s,
                'write_s': write_s,
                'cpu': row['process'].cpu_usage(),
                'memory': row['process'].mem_usage()
            })
        except Exception:
            pass
    return {'rows': final_backend_list}



import logging
import sys
from platform import python_version

from .postgres import Postgres
from .toolkit.app import BaseApplication
from .toolkit.configuration import OptionSpec
from .toolkit import validators as v
from .version import __version__


logger = logging.getLogger(__name__)


class Application(BaseApplication):
    # Agent specialisation of application.
    #
    # This include some defaults and Postgres connection.

    PROGRAM = "temboard-agent"
    VERSION = __version__
    REPORT_URL = "https://github.com/dalibo/temboard-agent"

    DEFAULT_CONFIGFILES = [
        '/etc/temboard-agent/temboard-agent.conf',
        'temboard-agent.conf',
    ]
    DEFAULT_PLUGINS = [
        "activity",
        "administration",
        "dashboard",
        "monitoring",
        "pgconf",
        "maintenance",
        "statements",
    ]

    def init_specs(self, app_specs):
        specs = super().init_specs(app_specs)

        def add_specs(*new_specs):
            for spec in new_specs:
                specs[str(spec)] = spec

        # These are *core* because they are needed to load plugins.
        s = 'postgresql'
        add_specs(
            OptionSpec(
                s, 'host', default='/var/run/postgresql', validator=v.dir_),
            OptionSpec(s, 'instance', default='main'),
            OptionSpec(s, 'port', default=5432, validator=v.port),
            OptionSpec(s, 'user', default='postgres'),
            OptionSpec(s, 'password'),
            OptionSpec(s, 'dbname', default='postgres'),
        )

        return specs

    def core_specs(self):
        for spec in super().core_specs():
            yield spec

        for name, spec in self.config_specs.items():
            if name.startswith('postgresql_'):
                yield spec

    def apply_config(self):
        self.postgres = Postgres(app=self, **self.config.postgresql)
        return super().apply_config()

    def check_compatibility(self, pg_version):
        # check for compatibility with plugins
        for name, plugin in self.plugins.items():
            if pg_version < plugin.PG_MIN_VERSION[0]:
                logger.error(
                    "%s plugin is incompatible with Postgres below %s",
                    name, plugin.PG_MIN_VERSION[1],
                )

    def log_versions(self):
        versions = inspect_versions()
        logger.info(
            "Running on %s %s.",
            versions['distname'], versions['distversion'])
        logger.info(
            "Using Python %s (%s).",
            versions['python'], versions['pythonbin'])
        logger.info(
            "Using libpq %s, Psycopg2 %s.",
            versions['libpq'], versions['psycopg2'],
        )


def inspect_versions():
    from .toolkit.versions import (
        format_pq_version,
        read_distinfo,
        read_libpq_version,
    )
    from psycopg2 import __version__ as psycopg2_version

    distinfos = read_distinfo()

    return dict(
        distname=distinfos['NAME'],
        distversion=distinfos['VERSION'],
        libpq=format_pq_version(read_libpq_version()),
        psycopg2=psycopg2_version,
        python=python_version(),
        pythonbin=sys.executable,
        temboard=__version__,
    )



# Various functions to expose components version

import ctypes
import re


def read_distinfo():  # pragma: nocover
    with open('/etc/os-release') as fo:
        distinfos = parse_lsb_release(fo)
    return distinfos


def parse_lsb_release(lines):
    _assignement_re = re.compile(
        r"""(?P<variable>[A-Z_]+)="(?P<value>[^"]+)"$"""
    )
    infos = dict()
    for line in lines:
        m = _assignement_re.match(line)
        if not m:
            continue
        infos[m.group('variable')] = m.group('value')
    return infos


def read_libpq_version():
    # Search libpq version bound to this process.

    try:
        # For psycopg2 2.7+
        from psycopg2.extensions import libpq_version
        return libpq_version()
    except ImportError:
        __import__('psycopg2')

        # Search for libpq.so path in loaded libraries.
        with open('/proc/self/maps') as fo:
            for line in fo:
                values = line.split()
                path = values[-1]
                if '/libpq' in path:
                    break
            else:  # pragma: nocover
                raise Exception("libpq.so not loaded")

        libpq = ctypes.cdll.LoadLibrary(path)
        return libpq.PQlibVersion()


def format_pq_version(version):
    pqnums = [
        version / 10000,
        version % 100,
    ]
    if version <= 100000:
        pqnums[1:1] = [(version % 10000) / 100]
    return '.'.join(str(int(n)) for n in pqnums)



def pg_version(conn):
    """
    Returns the PostgreSQL server version as numeric and full version.
    """
    num_version = conn.server_version
    full_version = conn.query_scalar("SELECT version()")
    return dict(numeric=num_version, full=full_version)



from ...routing import RouteSet

from . import functions as pgconf_functions
from .types import (
    T_PGSETTINGS_CATEGORY,
)


routes = RouteSet(prefix=b'/pgconf')


@routes.get(b'/configuration')
@routes.get(b'/configuration/category/' + T_PGSETTINGS_CATEGORY)
def get_pg_conf(http_context, app):
    with app.postgres.connect() as conn:
        return pgconf_functions.get_settings(conn, http_context)


@routes.get(b'/configuration/categories')
def get_pg_conf_categories(http_context, app):
    with app.postgres.connect() as conn:
        return pgconf_functions.get_settings_categories(conn)


@routes.post(b'/configuration')
def post_pg_conf(http_context, app):
    with app.postgres.connect() as conn:
        return pgconf_functions.post_settings(conn, app.config, http_context)


@routes.get(b'/configuration/status')
def get_pg_conf_status(http_context, app):
    with app.postgres.connect() as conn:
        return pgconf_functions.get_settings_status(conn)


class PgConfPlugin:
    PG_MIN_VERSION = (90500, 9.5)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        self.app.router.add(routes)

    def unload(self):
        self.app.router.remove(routes)
